from edustudio.utils.common import UnifyConfig
import os
from enum import Enum
import pandas as pd
from .base_datatpl import BaseDataTPL
import copy
from torch.utils.data import Dataset, DataLoader, default_collate
import pickle
import json
from deepdiff import DeepDiff
import importlib
import torch


class DataTPLMode(Enum):
    TRAIN=1
    VALID=2
    TEST=3
    MANAGER=4


class DataTPLStatus(object):
    def __init__(self, mode=DataTPLMode.MANAGER, fold_id=None) -> None:
        self.mode = mode
        self.fold_id = fold_id
        

class GeneralDataTPL(BaseDataTPL):
    """General Data Template
    """
    default_cfg = {
        'seperator': ',',
        'n_folds': 1,
        'is_dataset_divided': False,
        'is_save_cache': False,
        'cache_id': 'cache_default',
        'load_data_from': 'middata', # ['rawdata', 'middata', 'cachedata']
        'inter_exclude_feat_names': (),
        'raw2mid_op': "None", 
        'mid2cache_op_seq': []
    }

    def __init__(
                self, cfg:UnifyConfig,
                df: pd.DataFrame=None,
                df_train: pd.DataFrame=None,
                df_valid: pd.DataFrame=None,
                df_test: pd.DataFrame=None,
                status: DataTPLStatus=DataTPLStatus()
            ):
        super().__init__(cfg)
        self.df = df
        self.df_train = df_train
        self.df_valid = df_valid
        self.df_test = df_test
        self.status = status
        self.dict_main = None

        self.df_train_folds = []
        self.df_valid_folds = []
        self.df_test_folds = []
        self.dict_train_folds = []
        self.dict_valid_folds = []
        self.dict_test_folds = []

        # self.r2m_op = self._get_r2m_op()
        self.m2c_op_list = self._get_m2c_op_list()

        self.datatpl_cfg['dt_info'] = {}
        if self.datatpl_cfg['load_data_from'] == 'cachedata':
            self.load_cache()
            self.check_cache()
            self.process_data()
            self.logger.info(f"Load from cache successfully: {self.datatpl_cfg['cache_id']}")
            self.logger.info(self.datatpl_cfg['dt_info'])
        else:
            # self.feat2type = {}
            self.process_data()
            self.df2dict()
            if self.datatpl_cfg['is_save_cache']:
                self.save_cache()
            self.logger.info(self.datatpl_cfg['dt_info'])

    @classmethod
    def from_cfg(cls, cfg):
        """an interface to instantiate a data template

        Args:
            cfg (UnifyConfig): the global config object

        Returns:
           BaseDataTPL
        """
        if not os.path.exists(cfg.frame_cfg.data_folder_path) or len(os.listdir(cfg.frame_cfg.data_folder_path)) == 0:
            cls.download_dataset(cfg)
        
        load_data_from = cfg.datatpl_cfg['load_data_from']
        if load_data_from == 'middata':
            kwargs = cls.load_data(cfg)
        elif load_data_from == 'rawdata':
            r2m_op = cls._get_r2m_op(cfg)
            r2m_op.process() # 生成middata文件夹
            kwargs = cls.load_data(cfg) # 基于middata文件夹导入数据
        else:
            kwargs = {}

        return cls(cfg, **kwargs)
    
    @property
    def common_str2df(self):
        """get the common data object

        Returns:
            dict: the common data object through the atomic operations
        """
        return {
            "df": self.df, "df_train": self.df_train, "df_valid": self.df_valid,
            "df_test": self.df_test, "dt_info": self.datatpl_cfg['dt_info']
        }
    
    def process_load_data_from_middata(self):
        """process middata
        """
        kwargs = self.common_str2df
        for op in self.m2c_op_list:
            self.logger.info(f"Processing mid2cache op: {op.__class__.__name__}")
            kwargs = op.process(**kwargs)
            assert kwargs is not None
            op.set_dt_info(**kwargs)
        self.logger.info(f"Mid2cache op done!")
        for k,v in kwargs.items():
            setattr(self, k, v)

        self.final_kwargs = kwargs

        if self.datatpl_cfg['is_dataset_divided']:
            raise NotImplementedError # 统计数据集

    def process_data(self):
        """process data
        """
        load_data_from = self.datatpl_cfg['load_data_from']
        if load_data_from != 'cachedata':
            self.process_load_data_from_middata()
    
    @classmethod
    def load_data(cls, cfg): # 只在middata存在时调用
        """load data from disk
        """
        is_dataset_divided = cfg.datatpl_cfg['is_dataset_divided']
        if cfg.datatpl_cfg['n_folds'] == 1:
            kwargs = cls.load_data_from_undivided(cfg) # 1折时，如果数据没划分好，那么直接导入
        else:
            if is_dataset_divided: # 五折时，不支持已经划分好的
                raise ValueError("In the setting of n_fold>1, unsupport is_dataset_divided=True")
            else:
                kwargs = cls.load_data_from_undivided(cfg)
        for df in kwargs.values(): cls._preprocess_feat(df) # 类型转换
        return kwargs

    @classmethod
    def load_data_from_undivided(cls, cfg):
        """load undivided data from disk
        """
        fph = f'{cfg.frame_cfg.data_folder_path}/middata/{cfg.dataset}.inter.csv'
        
        exclude_feats = cfg.datatpl_cfg['inter_exclude_feat_names']
        assert len(set(exclude_feats) & {'stu_id:token', 'exer_id:token', 'label:float'}) == 0

        sep = cfg.datatpl_cfg['seperator']
        df = cls._load_atomic_csv(fph, exclude_headers=exclude_feats, sep=sep)

        return {"df": df}
    
    @classmethod
    def load_data_from_divided(cls, cfg):
        """load divided data from disk
        """
        train_fph = f'{cfg.frame_cfg.data_folder_path}/middata/{cfg.dataset}.train.inter.csv'
        valid_fph = f'{cfg.frame_cfg.data_folder_path}/middata/{cfg.dataset}.valid.inter.csv'
        test_fph = f'{cfg.frame_cfg.data_folder_path}/middata/{cfg.dataset}.test.inter.csv'

        exclude_feats = cfg.datatpl_cfg['inter_exclude_feat_names']
        assert len(set(exclude_feats) & {'stu_id:token', 'exer_id:token', 'label:float'}) == 0

        sep = cfg.datatpl_cfg['seperator']
        df_train = cls._load_atomic_csv(train_fph, exclude_headers=exclude_feats, sep=sep)
        df_test = cls._load_atomic_csv(test_fph, exclude_headers=exclude_feats, sep=sep)
        df_valid = None
        if os.path.exists(valid_fph):
            df_valid = cls._load_atomic_csv(valid_fph, exclude_headers=exclude_feats, sep=sep)

        return {"df_train": df_train, "df_valid": df_valid, "df_test": df_test}

    @staticmethod
    def _load_atomic_csv(fph, exclude_headers, sep=','):
        headers = pd.read_csv(fph, nrows=0).columns.tolist()
        df = pd.read_csv(fph, sep=sep, encoding='utf-8', usecols=set(headers) - set(exclude_headers))
        return df

    def build_dataloaders(self):
        """build dataloaders that would be convey to training template
        """
        batch_size = self.traintpl_cfg['batch_size']
        num_workers = self.traintpl_cfg['num_workers']
        eval_batch_size = self.traintpl_cfg['eval_batch_size']
        train_dt_list, valid_dt_list, test_dt_list = self.build_datasets()
        train_loader_list, valid_loader_list, test_loader_list = [], [], []

        for fid in range(self.datatpl_cfg['n_folds']):
            train_loader = DataLoader(dataset=train_dt_list[fid], shuffle=True, batch_size=batch_size, num_workers=num_workers, collate_fn=self.collate_fn)
            train_loader_list.append(train_loader)
            if self.hasValidDataset:
                valid_loader = DataLoader(dataset=valid_dt_list[fid], shuffle=False, batch_size=eval_batch_size, num_workers=num_workers, collate_fn=self.collate_fn)
                valid_loader_list.append(valid_loader)
            test_loader = DataLoader(dataset=test_dt_list[fid], shuffle=False, batch_size=eval_batch_size, num_workers=num_workers, collate_fn=self.collate_fn)
            test_loader_list.append(test_loader)
        
        return train_loader_list, valid_loader_list, test_loader_list

    def save_cache(self):
        """save cache data
        """
        # chech path
        if os.path.exists(self.cache_folder_path):
            raise ValueError(f"cache_fold({self.cache_folder_path}) already exists, won't save cache")
        else:
            os.makedirs(self.cache_folder_path)

        # save cache
        train_folds_fph = f"{self.cache_folder_path}/dict_train_folds.pkl"
        valid_folds_fph = f"{self.cache_folder_path}/dict_valid_folds.pkl"
        test_folds_fph = f"{self.cache_folder_path}/dict_test_folds.pkl"
        final_kwargs_fph = f"{self.cache_folder_path}/final_kwargs.pkl"

        self.save_pickle(train_folds_fph, self.dict_train_folds)
        self.save_pickle(valid_folds_fph, self.dict_valid_folds)
        self.save_pickle(test_folds_fph, self.dict_test_folds)
        self.save_pickle(final_kwargs_fph, self.final_kwargs)

        with open(f"{self.cache_folder_path}/datatpl_cfg.json", 'w', encoding='utf-8') as f:
            json.dump(json.loads(self.datatpl_cfg.dump_fmt()), fp=f, indent=2, ensure_ascii=False)

    def check_cache(self):
        """check whether the cache data is consistent with current config
        """
        with open(f"{self.cache_folder_path}/datatpl_cfg.json", 'r', encoding='utf-8') as f:
            cache_datatpl_cfg = json.load(f)
        
        temp_cache_datatpl_cfg = copy.deepcopy(cache_datatpl_cfg)
        del temp_cache_datatpl_cfg['dt_info']
        del temp_cache_datatpl_cfg['load_data_from']
        if 'is_save_cache' in temp_cache_datatpl_cfg:
            del temp_cache_datatpl_cfg['is_save_cache']
        # del temp_cache_datatpl_cfg['raw2mid_op']
        # del temp_cache_datatpl_cfg['mid2cache_op_seq']
        curr_datatpl_cfg = copy.deepcopy(json.loads(self.datatpl_cfg.dump_fmt()))
        del curr_datatpl_cfg['dt_info']
        del curr_datatpl_cfg['load_data_from']
        if 'is_save_cache' in curr_datatpl_cfg:
            del curr_datatpl_cfg['is_save_cache']
        # del curr_datatpl_cfg['raw2mid_op']
        # del curr_datatpl_cfg['mid2cache_op_seq']
        diff = DeepDiff(temp_cache_datatpl_cfg, curr_datatpl_cfg)

        if len(diff) != 0:
            raise ValueError(f"check cache error: {diff}")
        
        self.datatpl_cfg['dt_info'] = cache_datatpl_cfg['dt_info'] # partial load_cache

    def load_cache(self):
        """load cache data from disk
        """
        if self.datatpl_cfg['load_data_from'] != 'cachedata':
            return
        if not os.path.exists(self.cache_folder_path):
            raise ValueError(f"cache_fold({self.cache_folder_path}) not exists, can't load cache")
        
        train_folds_fph = f"{self.cache_folder_path}/dict_train_folds.pkl"
        valid_folds_fph = f"{self.cache_folder_path}/dict_valid_folds.pkl"
        test_folds_fph = f"{self.cache_folder_path}/dict_test_folds.pkl"
        final_kwargs_fph = f"{self.cache_folder_path}/final_kwargs.pkl"

        self.dict_train_folds = self.load_pickle(train_folds_fph)
        self.dict_valid_folds = self.load_pickle(valid_folds_fph)
        self.dict_test_folds = self.load_pickle(test_folds_fph)
        self.final_kwargs = self.load_pickle(final_kwargs_fph)

        for k,v in self.final_kwargs.items():
            if not hasattr(self, k):
                setattr(self, k, v)
                self.logger.info(f"[load cache] set {k} from final_kwargs to current data template")
            else:
                self.logger.info(f"[load cache] duplicated attribute in final_kwargs: {k}")

    def build_datasets(self):
        """build datasets
        """
        n_folds = self.datatpl_cfg['n_folds']
        assert len(self.dict_train_folds) == n_folds
        assert self.status.mode == DataTPLMode.MANAGER

        train_dt_list, valid_dt_list, test_dt_list = [], [], []

        for fid in range(n_folds):
            train_dataset = self._copy()
            train_dataset.set_mode(DataTPLMode.TRAIN, fid)
            train_dt_list.append(train_dataset)

            valid_dataset = None
            if self.hasValidDataset:
                valid_dataset = self._copy()
                valid_dataset.set_mode(DataTPLMode.VALID, fid)
                valid_dt_list.append(valid_dataset)
            
            test_dataset = self._copy()
            test_dataset.set_mode(DataTPLMode.TEST, fid)
            test_dt_list.append(test_dataset)
        
        return train_dt_list, valid_dt_list, test_dt_list

    @staticmethod
    def collate_fn(batch):
        return default_collate(batch)
    
    def set_mode(self, mode: DataTPLMode, fold_id):
        """set mode of current data template

        Args:
            mode (DataTPLMode): mode
            fold_id (int): id of fold
        """
        self.status.mode = mode
        self.status.fold_id = fold_id
        if mode is DataTPLMode.MANAGER:
            self._set_mode_manager()
        elif mode is DataTPLMode.TRAIN:
            self._set_mode_train()
        elif mode is DataTPLMode.VALID:
            self._set_mode_valid()
        elif mode is DataTPLMode.TEST:
            self._set_mode_test()
        else:
            raise ValueError(f"unknown type of mode:{mode}")

        self.length = next(iter(self.dict_main.values())).shape[0]

    def _set_mode_manager(self):
        """progress of manager mode
        """
        self.dict_main = None
        self.status.mode = DataTPLMode.MANAGER
        self.status.fold_id = None

    def _set_mode_train(self):
        """progress of train mode
        """
        self.dict_main = self.dict_train_folds[self.status.fold_id]
    
    def _set_mode_valid(self):
        """progress of valid mode
        """
        self.dict_main = self.dict_valid_folds[self.status.fold_id]
    
    def _set_mode_test(self):
        """progress of test mode
        """
        self.dict_main = self.dict_test_folds[self.status.fold_id]

    def _check_params(self):
        """check validation of default parameters
        """
        super()._check_params()
        assert self.datatpl_cfg['load_data_from'] in {'rawdata', 'middata', 'cachedata'}
        assert 'dt_info' not in self.datatpl_cfg
        assert self.datatpl_cfg['n_folds'] > 0

    def __len__(self):
        return self.length
    
    def __getitem__(self, index):
        return {
            k: v[index] for k,v in self.dict_main.items()
        }
    
    def _copy(self):
        """copy current instance
        """
        obj = super()._copy()
        obj.status = copy.copy(obj.status)
        return obj
    
    @property
    def hasValidDataset(self):
        """whether exists validation dataset
        """
        return self.dict_valid_folds is not None and len(self.dict_valid_folds) > 0
    
    @property
    def cache_folder_path(self):
        """folder path of cache data
        """
        save_cache_id = self.datatpl_cfg['cache_id']
        return f"{self.frame_cfg.data_folder_path}/cachedata/{self.cfg.dataset}_{save_cache_id}/"
    
    def save_pickle(self, filepath, obj):
        """save data into pickle file
        """
        with open(filepath, 'wb') as fb:
            pickle.dump(obj, fb, protocol=pickle.HIGHEST_PROTOCOL)
    
    def load_pickle(self, filepath):
        """load data into pickle file
        """
        with open(filepath, 'rb') as fb:
            return pickle.load(fb)

    @classmethod
    def get_default_cfg(cls, mid2cache_op_seq, **kwargs):
        """Get final default config object

        Args:
            mid2cache_op_seq (List[Union[BaseMid2Cache,str]]): Mid2Cahce Sequence

        Returns:
            UnifyConfig: the final default config object
        """
        cfg = UnifyConfig()
        for _cls in cls.__mro__:
            if not hasattr(_cls, 'default_cfg'):
                break
            cfg.update(_cls.default_cfg, update_unknown_key_only=True)

        # 获取mid_cache_op_seq
        mid2cache_op_seq = None
        if mid2cache_op_seq is None:
            for _cls in cls.__mro__:
                if not hasattr(_cls, 'default_cfg'):
                    break
                mid2cache_op_seq = _cls.default_cfg.get('mid2cache_op_seq', None)
                if mid2cache_op_seq is not None: break
            assert mid2cache_op_seq is not None
        
        # cfg.mid2cache_op_seq = mid2cache_op_seq
        from edustudio.atom_op.mid2cache import BaseMid2Cache
        from edustudio.atom_op.raw2mid import BaseRaw2Mid
        for op in mid2cache_op_seq:
            if isinstance(op, str):
                op = importlib.import_module('edustudio.atom_op.mid2cache').__getattribute__(op)
            elif issubclass(op, BaseMid2Cache):
                pass
            else:
                raise ValueError(f"Unknown type of Mid2Cache: {op}")
            cfg[op.__name__] = op.get_default_cfg()

        return cfg

    @classmethod
    def _get_r2m_op(cls, cfg):
        """Get Raw2Mid operation

        Args:
            cfg (UnifyConfig): the global config object

        Returns:
            BaseRaw2Mid: Raw2Mid operation
        """
        from edustudio.atom_op.raw2mid import BaseRaw2Mid
        r2m_op = cfg.datatpl_cfg['raw2mid_op']
        assert r2m_op is not None or r2m_op != "None"
        if isinstance(r2m_op, str):
            r2m_op = importlib.import_module('edustudio.atom_op.raw2mid').__getattribute__(r2m_op)
        elif issubclass(r2m_op, BaseRaw2Mid):
            pass
        else:
            raise ValueError(f"unsupported raw2mid_op:{r2m_op}")
        return r2m_op.from_cfg(cfg)
    
    def _get_m2c_op_list(self):
        """Get Mid2Cache operation sequence

        Returns:
           List[BaseMid2Cache]: Mid2Cache operation sequence
        """
        from edustudio.atom_op.mid2cache import BaseMid2Cache
        m2c_op_list = self.datatpl_cfg['mid2cache_op_seq']
        op_list = []
        for op in m2c_op_list:
            if isinstance(op, str):
                op = importlib.import_module('edustudio.atom_op.mid2cache').__getattribute__(op)
            elif issubclass(op, BaseMid2Cache):
                pass
            else:
                raise ValueError(f"unsupported mid2cache_op:{op}")
            op_list.append(op.from_cfg(self.cfg))
        return op_list


    def df2dict(self):
        """convert dataframe object into dict
        """
        # if self.datatpl_cfg['is_dataset_divided'] is True:
        #     self.df_train_folds.append(self.df_train)
        #     self.df_valid_folds.append(self.df_valid)
        #     self.df_test_folds.append(self.df_test)
        
        for tmp_df in self.df_train_folds:
            self.dict_train_folds.append(self._df2dict(tmp_df))
            
        if len(self.df_valid_folds) > 0:
            for tmp_df in self.df_valid_folds:
                self.dict_valid_folds.append(self._df2dict(tmp_df))
            
        for tmp_df in self.df_test_folds:
            self.dict_test_folds.append(self._df2dict(tmp_df))

    @staticmethod
    def _df2dict(dic_raw):
        """convert dataframe into dict
        """
        dic = {}
        for k in dic_raw:
            if type(dic_raw) is not dict:
                v = torch.from_numpy(dic_raw[k].to_numpy())
            else:
                v = dic_raw[k]
            if ":" in k:
                k = k.split(":")[0]
            dic[k] = v
        return dic

    def get_extra_data(self, **kwargs):
        """an interface to construct extra data except the data from forward API
        """
        extra_data = super().get_extra_data(**kwargs)
        # extra_data.update({
        #     'df': self.df,
        #     'df_train': self.df_train,
        #     'df_valid': self.df_valid,
        #     'df_test': self.df_test
        # })
        extra_data.update(self.final_kwargs)
        return extra_data
    
    @staticmethod
    def _preprocess_feat(df):
        """convert data format after loading files according to field type

        Args:
            df (DataFrame): data
        """
        for col in df.columns:
            _split = col.split(":")
            assert len(_split) == 2 or len(_split) == 1, f"col {col} is not valid"
            if len(_split) == 1: continue
            _, col_type = col.split(":")
            if col_type == 'float':
                df[col] = df[col].astype('float32')
            elif col_type == 'token_seq':
                try:
                    if type(df[col].iloc[0]) in {np.ndarray, list}:
                        continue
                    df[col] = df[col].astype(str).apply(lambda x: eval(x))
                except:
                    df[col] = df[col].astype(str).apply(lambda x: [i for i in x.split(",")])
            elif col_type == 'float_seq':
                try:
                    df[col] = df[col].astype(str).apply(lambda x: eval(x))
                except:
                    df[col] = df[col].astype(str).apply(lambda x: [float(i) for i in x.split(",")])
            elif col_type == 'int_seq':
                try:
                    df[col] = df[col].astype(str).apply(lambda x: eval(x))
                except:
                    df[col] = df[col].astype(str).apply(lambda x: [int(i) for i in x.split(",")])
            elif col_type == 'str_seq':
                try:
                    df[col] = df[col].astype(str).apply(lambda x: eval(x))
                except:
                    df[col] = df[col].astype(str).apply(lambda x: [str(i) for i in x.split(",")])
            elif col_type == 'str' or col_type == 'string':
                df[col] = df[col].astype(str)
            elif col_type =='int':
                df[col] = df[col].astype(int)
            
    @staticmethod
    def _unwrap_feat(df:pd.DataFrame):
        """unwrap the type of field

        Args:
            df (pd.DataFrame): dataframe after unwrapping
        """
        for col in df.columns:
            col_name, col_type = col.split(":")
            df.rename(columns={col:col_name}, inplace=True)
        
    @classmethod
    def get_default_cfg(cls, **kwargs):
        """Get the final default config object

        Returns:
            UnifyConfig: the final default config object
        """
        cfg = UnifyConfig()
        for _cls in cls.__mro__:
            if not hasattr(_cls, 'default_cfg'):
                break
            cfg.update(_cls.default_cfg, update_unknown_key_only=True)
        
        for op in kwargs.get('mid2cache_op_seq', None) or cls.default_cfg['mid2cache_op_seq']:
            if isinstance(op, str):
                op = importlib.import_module('edustudio.atom_op.mid2cache').__getattribute__(op)
            if op.__name__ not in cfg:
                cfg[op.__name__] = op.get_default_cfg()
            else:
                cfg[op.__name__] = UnifyConfig(cfg[op.__name__])
                cfg[op.__name__].update(op.get_default_cfg(), update_unknown_key_only=True)
        return cfg

    
    def set_info_for_fold(self, fold_id):
        """Get data information when a specifying fold id

        Args:
            fold_id (int): id of fold
        """
        pass



from ..common import BaseMid2Cache
from itertools import chain
import torch
import pandas as pd
import numpy as np
from edustudio.datatpl.utils import PadSeqUtil
import copy


class M2C_EERNN_OP(BaseMid2Cache):
    default_cfg = {
        'word_emb_dim': 50,
        'max_sentence_len': 100,
    }

    def process(self, **kwargs):
        dt_info = kwargs['dt_info']
        train_dict_list = kwargs['df_train_folds']
        df_exer = kwargs['df_exer']
        assert len(train_dict_list) > 0
        import gensim

        word_emb_dict_list,content_mat_list = [], []
        for train_dict in train_dict_list:
            exers = np.unique(train_dict['exer_seq:token_seq'])
            sentences = df_exer[df_exer['exer_id:token'].isin(exers)]['content:token_seq'].tolist()
            # desc_dict = gensim.corpora.Dictionary(sentences)
            # word_set = list(desc_dict.token2id.keys())
            # word2id = {w:i+1 for i,w in enumerate(word_set)}
            model = gensim.models.word2vec.Word2Vec(
                sentences, vector_size=self.m2c_cfg['word_emb_dim'],
            )
            wv_from_bin = model.wv

            word2id = copy.deepcopy(model.wv.key_to_index)
            word2id = {k: word2id[k] + 1 for k in word2id}
            word_emb_dict = {word2id[key]: wv_from_bin[key] for key in word2id}
            word_emb_dict[0] = np.zeros(shape=(self.m2c_cfg['word_emb_dim'], ))

            # 将训练集、验证集、测试集中未出现在word_emb中的单词，全部替换成ID为0，并进行padding
            df_exer['content:token_seq'] = df_exer['content:token_seq'].apply(lambda x: [word2id.get(xx, 0) for xx in x])        
            pad_mat, _, _ = PadSeqUtil.pad_sequence(
                df_exer['content:token_seq'].tolist(), maxlen=self.m2c_cfg['max_sentence_len'], padding='post',
                is_truncate=True, truncating='post', value=0, 
            )
            df_exer['content:token_seq'] = [pad_mat[i].tolist() for i in range(pad_mat.shape[0])]
            
            tmp_df_Q = df_exer.set_index('exer_id:token')
            content_mat = torch.from_numpy(np.vstack(
                [tmp_df_Q.loc[exer_id]['content:token_seq'] for exer_id in range(dt_info['exer_count'])]
            ))

            word_emb_dict_list.append(word_emb_dict)
            content_mat_list.append(content_mat)
        kwargs['word_emb_dict_list'] = word_emb_dict_list
        kwargs['content_mat_list'] = content_mat_list
        return kwargs

    def set_dt_info(self, dt_info, **kwargs):
        dt_info['word_emb_dim'] = self.m2c_cfg['word_emb_dim']



import numpy as np

from ..common import EduDataTPL

class CL4KTDataTPL(EduDataTPL):
    default_cfg = {
        'mid2cache_op_seq': ['M2C_Label2Int', 'M2C_ReMapId', 'M2C_GenUnFoldKCSeq', 'M2C_CL4KT_OP', 'M2C_RandomDataSplit4KT'],
        'M2C_CL4KT_OP': {
            'sequence_truncation': 'recent',
        }
    }

    def __getitem__(self, index):
        dic = super().__getitem__(index)

        return dic



from .raw2mid import BaseRaw2Mid
import pandas as pd


class R2M_ASSIST_1213(BaseRaw2Mid):
    def process(self, **kwargs):
        use_cols = ["user_id", "skill_id", "start_time", "problem_id", "correct", "ms_first_response", 'assignment_id',
                    'attempt_count', 'student_class_id', 'teacher_id']
        df = pd.read_csv(f"{self.rawpath}/2012-2013-data-with-predictions-4-final.csv", low_memory=False,
                         usecols=use_cols)
        df = df.dropna(subset=use_cols)
        df['correct'] = df['correct'].apply(int)
        df['skill_id'] = df['skill_id'].apply(int)
        df = df[df['correct'].isin([0, 1])]  # filter responses

        from datetime import datetime
        def change2timestamp(t, hasf=True):
            if hasf:
                timeStamp = datetime.strptime(t, "%Y-%m-%d %H:%M:%S.%f").timestamp() * 1000
            else:
                timeStamp = datetime.strptime(t, "%Y-%m-%d %H:%M:%S").timestamp() * 1000
            return int(timeStamp)

        df['start_timestamp'] = df['start_time'].apply(lambda x: change2timestamp(x, hasf='.' in x))
        # df['skill_id'] = df['skill_id'].apply(lambda x:[x])
        df.rename(columns={'problem_id': 'exer_id', 'correct': 'label', 'ms_first_response': 'cost_time',
                           'skill_id': 'cpt_seq', 'student_class_id': 'class_id'}, inplace=True)
        df.drop(columns=['start_time'], inplace=True)

        grouped = df.groupby('user_id', as_index=False)

        # 定义一个函数来修改每个group的DataFrame
        def modify_group(group):
            group.sort_values(['start_timestamp'], inplace=True)
            group['order'] = range(len(group))
            return group

        # 使用apply函数来应用修改函数到每个group的DataFrame
        df_modified = grouped.apply(modify_group)
        df = df_modified.reset_index(drop=True)
        df_inter = df[['exer_id', 'user_id', 'start_timestamp', 'order', 'cost_time', 'label']]
        df_inter = df_inter.rename(
            columns={'user_id': 'stu_id:token', 'exer_id': 'exer_id:token', 'label': 'label:float',
                     'start_timestamp': 'start_timestamp:float', 'cost_time': 'cost_time:float',
                     'order': 'order_id:float'})
        df_exer = df[['exer_id', 'cpt_seq', 'assignment_id']]
        df_exer = df_exer.drop_duplicates(subset=['exer_id'])
        df_exer = df_exer.rename(
            columns={'exer_id': 'exer_id:token', 'assignment_id': 'assignment_id:token',
                     'cpt_seq': 'cpt_seq:token_seq'})
        df_stu = df[['user_id', 'class_id', 'teacher_id']]
        df_stu = df_stu.drop_duplicates(subset=['user_id'])
        df_stu = df_stu.rename(
            columns={'user_id': 'stu_id:token', 'class_id': 'class_id:token',
                     'teacher_id': 'teacher_id:token'})
        cpt_ls = []
        df['cpt_seq'].astype(str).apply(lambda x: cpt_ls.extend(x.split(',')))
        # ser_cpt = df['cpt_seq'].astype(str).apply(lambda x: [int(i) for i in x.split(',')])
        # ser_cpt.to_list()
        # cpt_set = set(cpt_ls)
        # cfg.exer_count = len(df['exer_id'].unique())
        # cfg.stu_count = len(df['user_id'].unique())
        # cfg.cpt_count = len(cpt_set)
        # cfg.interaction_count = len(df_inter)
        df_inter.to_csv(f"{self.midpath}/{self.dt}.inter.csv", index=False, encoding='utf-8')
        df_stu.to_csv(f"{self.midpath}/{self.dt}.stu.csv", index=False, encoding='utf-8')
        df_exer.to_csv(f"{self.midpath}/{self.dt}.exer.csv", index=False, encoding='utf-8')

    


import networkx as nx
from ..common import BaseMid2Cache
import torch
from torch.nn import functional as F
import numpy as np


class M2C_QDKT_OP(BaseMid2Cache):
    default_cfg = {}

    def process(self, **kwargs):
        self.df_Q = kwargs['df_exer']
        dt_info = kwargs['dt_info']
        self.num_q = dt_info['exer_count']
        self.num_c = dt_info['cpt_count']
        self.Q_mat = kwargs['Q_mat']
        laplacian_matrix = self.laplacian_matrix_by_vectorization()
        kwargs['laplacian_matrix'] = laplacian_matrix
        return kwargs

    def laplacian_matrix_by_vectorization(self):
        normQ = F.normalize(self.Q_mat.float(), p=2, dim=-1)
        A = torch.mm(normQ, normQ.T) > (1 - 1/len(normQ))
        A = A.int()  #Adjacency matrix
        D = A.sum(-1, dtype=torch.int32)
        diag_idx = [range(len(A)), range(len(A))]
        A[diag_idx] = D - A[diag_idx]
        return A

    def generate_graph(self):

        graph = nx.Graph()
        len1 = len(self.Q_mat)
        graph.add_nodes_from([i for i in range(1, len1 + 1)])
        for index in range(len1 - 1):
            for bindex in range(index + 1, len1):
                if not (False in (self.Q_mat[index, :] == self.Q_mat[bindex, :]).tolist()):
                    graph.add_edge(index + 1, bindex + 1)
        return graph

        # 求图的拉普拉斯矩阵 L = D - A

    def laplacian_matrix(self, graph):
        # 求邻接矩阵
        A = np.array(nx.adjacency_matrix(graph).todense())
        A = -A
        for i in range(len(A)):
            # 求顶点的度
            degree_i = graph.degree(i + 1)  # 节点编号从1开始，若从0开始，将i+1改为i
            A[i][i] = A[i][i] + degree_i
        return A



from ..common import EduDataTPL


class MGCDDataTPL(EduDataTPL):
    default_cfg = {
        'mid2cache_op_seq': ['M2C_Label2Int', 'M2C_FilterRecords4CD', 'M2C_MGCD_OP', 'M2C_ReMapId', 'M2C_RandomDataSplit4CD', 'M2C_GenQMat'],
    }

    def get_extra_data(self):
        dic = super().get_extra_data()
        dic.update({
            'inter_student': self.final_kwargs['df_inter_stu'],
            'df_G': self.final_kwargs['df_stu']
        })
        return dic

    def df2dict(self):
        super().df2dict()
        self._unwrap_feat(self.final_kwargs['df_inter_stu'])
        self._unwrap_feat(self.final_kwargs['df_stu'])



from ..common import EduDataTPL

class CNCDFDataTPL(EduDataTPL):
    default_cfg = {
        'mid2cache_op_seq': ['M2C_Label2Int', 'M2C_FilterRecords4CD', 'M2C_ReMapId', 'M2C_RandomDataSplit4CD', 'M2C_GenQMat'],
    }

    def get_extra_data(self):
        return {
            'content': self.df_exer['content:token_seq'].to_list(),
            'Q_mat': self.final_kwargs['Q_mat']
        }




from .EERNNDataTPL import EERNNDataTPL
import numpy as np


class EKTDataTPL(EERNNDataTPL):
    default_cfg = {
        'mid2cache_op_seq': ['M2C_Label2Int', 'M2C_ReMapId', 'M2C_BuildSeqInterFeats', 'M2C_RandomDataSplit4KT', 'M2C_GenKCSeq', 'M2C_EERNN_OP'],
    }

    def __getitem__(self, index):
        dic = super().__getitem__(index)
        dic['cpt_seq'] = np.stack(
            [self.cpt_seq_padding[exer_seq] for exer_seq in dic['exer_seq']], axis=0
        )
        dic['cpt_seq_mask'] = np.stack(
            [self.cpt_seq_mask[exer_seq] for exer_seq in dic['exer_seq']], axis=0
        )
        return dic



r"""
DKTForget
##########################################

Reference:
    Koki Nagatani et al. "Augmenting Knowledge Tracing by Considering Forgetting Behavior" in WWW 2019.

Reference Code:
    https://github.com/pykt-team/pykt-toolkit/blob/main/pykt/models/dkt_forget.py

"""

from ..gd_basemodel import GDBaseModel
import torch.nn as nn
import torch
import torch.nn.functional as F


class DKTForget(GDBaseModel):
    default_cfg = {
        'emb_size': 100,
        'num_layers': 1,
        'dropout_rate': 0.2,
        'rnn_or_lstm': 'lstm',
        'integration_type': 'concat_multiply'
    }

    def build_cfg(self):
        self.n_user = self.datatpl_cfg['dt_info']['stu_count']
        self.n_item = self.datatpl_cfg['dt_info']['exer_count']
        self.n_rgap = self.datatpl_cfg['dt_info']['n_rgap']
        self.n_sgap = self.datatpl_cfg['dt_info']['n_sgap']
        self.n_pcount = self.datatpl_cfg['dt_info']['n_pcount']
        assert self.modeltpl_cfg['rnn_or_lstm'] in {'rnn', 'lstm'}
        
    def build_model(self):
        self.exer_emb = nn.Embedding(
            self.n_item * 2, self.modeltpl_cfg['emb_size']
        )
        self.integration_compont = IntegrationComponent(
            n_rgap=self.n_rgap, n_sgap=self.n_sgap, n_pcount=self.n_pcount,
            emb_dim=self.modeltpl_cfg['emb_size'], device=self.device,
            integration_type=self.modeltpl_cfg['integration_type']
        )
        if self.modeltpl_cfg['rnn_or_lstm'] == 'rnn':
            self.seq_model = nn.RNN(
                self.integration_compont.output_dim, self.modeltpl_cfg['emb_size'], 
                self.modeltpl_cfg['num_layers'], batch_first=True
            )
        else:
            self.seq_model = nn.LSTM(
                self.integration_compont.output_dim, self.modeltpl_cfg['emb_size'], 
                self.modeltpl_cfg['num_layers'], batch_first=True
            )
        self.dropout_layer = nn.Dropout(self.modeltpl_cfg['dropout_rate'])
        self.fc_layer = nn.Linear(self.integration_compont.output_dim, self.n_item)

    def forward(self, exer_seq, label_seq, r_gap, s_gap, p_count, **kwargs):
        input_x = self.exer_emb(exer_seq + label_seq.long() * self.n_item)
        input_x = self.integration_compont(input_x, r_gap, s_gap, p_count)
        h, _ = self.seq_model(input_x)
        output = self.integration_compont(h, r_gap, s_gap, p_count)
        output = self.dropout_layer(output)
        y_pd = self.fc_layer(output).sigmoid()
        return y_pd

    @torch.no_grad()
    def predict(self, **kwargs):
        y_pd = self(**kwargs)
        y_pd = y_pd[:, :-1].gather(
            index=kwargs['exer_seq'][:, 1:].unsqueeze(dim=-1), dim=2
        ).squeeze(dim=-1)
        y_pd = y_pd[kwargs['mask_seq'][:, 1:] == 1]
        y_gt = None
        if kwargs.get('label_seq', None) is not None:
            y_gt = kwargs['label_seq'][:, 1:]
            y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        return {
            'y_pd': y_pd,
            'y_gt': y_gt
        }

    def get_main_loss(self, **kwargs):
        y_pd = self(**kwargs)
        y_pd = y_pd[:, :-1].gather(
            index=kwargs['exer_seq'][:, 1:].unsqueeze(dim=-1), dim=2
        ).squeeze(dim=-1)
        y_pd = y_pd[kwargs['mask_seq'][:, 1:] == 1]
        y_gt = kwargs['label_seq'][:, 1:]
        y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        loss = F.binary_cross_entropy(
            input=y_pd, target=y_gt
        )
        return {
            'loss_main': loss
        }

    def get_loss_dict(self, **kwargs):
        return self.get_main_loss(**kwargs)


class IntegrationComponent(nn.Module):
    def __init__(self, n_rgap, n_sgap, n_pcount, emb_dim, integration_type="concat_multiply", device="cpu") -> None:
        super().__init__()
        self.rgap_eye = torch.eye(n_rgap).to(device)
        self.sgap_eye = torch.eye(n_sgap).to(device)
        self.pcount_eye = torch.eye(n_pcount).to(device)
        n_total = n_rgap + n_sgap + n_pcount

        if integration_type == 'concat':
            self.integration_func = self.concat
            self.output_dim = emb_dim + n_total
        elif integration_type == 'multiply':
            self.C = nn.Linear(n_total, emb_dim, bias=False)
            self.integration_func = self.multiply
            self.output_dim = emb_dim
        elif integration_type == 'concat_multiply':
            self.C = nn.Linear(n_total, emb_dim, bias=False)
            self.integration_func = self.concat_multiply
            self.output_dim = emb_dim + n_total
        elif integration_type == 'bi_interaction':
            self.C = nn.Linear(n_total, emb_dim, bias=False)
            self.integration_func = self.bi_interaction
            pass
        else:
            raise ValueError(f"unknown intefration_type: {integration_type}")

    def forward(self, v_t, r_gap, s_gap, p_count):
        return self.integration_func(v_t,r_gap, s_gap, p_count)

    def concat(self, v_t, r_gap, s_gap, p_count):
        rgap, sgap, pcount = self.rgap_eye[r_gap], self.sgap_eye[s_gap], self.pcount_eye[p_count]
        c_t = torch.cat((rgap, sgap, pcount), -1)
        return torch.cat((v_t, c_t), -1)

    def multiply(self, v_t, r_gap, s_gap, p_count):
        rgap, sgap, pcount = self.rgap_eye[r_gap], self.sgap_eye[s_gap], self.pcount_eye[p_count]
        c_t = torch.cat((rgap, sgap, pcount), -1)
        Cct = self.C(c_t)
        theta = torch.mul(v_t, Cct)
        return theta

    def concat_multiply(self, v_t, r_gap, s_gap, p_count):
        rgap = self.rgap_eye[r_gap]
        sgap = self.sgap_eye[s_gap]
        pcount = self.pcount_eye[p_count]
        c_t = torch.cat((rgap, sgap, pcount), -1)
        Cct = self.C(c_t)
        theta = torch.mul(v_t, Cct)
        theta = torch.cat((theta, c_t), -1)
        return theta

    def bi_interaction(self, v_t, r_gap, s_gap, p_count):
        raise NotImplementedError



from ..common import BaseMid2Cache
import numpy as np
import torch


class M2C_DKTDSC_OP(BaseMid2Cache):
    default_cfg = {
        'n_cluster': 7,
        'centroids_update_epoch': 2,
        'seg_length': 20
    }

    def process(self, **kwargs):
        dt_info = kwargs['dt_info']
        dt_info['n_cluster'] = self.m2c_cfg['n_cluster']
        self.seg_length = self.m2c_cfg['seg_length']
        self.stu_count = dt_info['stu_count']
        self.exer_count = dt_info['exer_count']
        self.num_cluster = self.m2c_cfg['n_cluster']
        self.window_size = dt_info['real_window_size']

        df_train_folds = kwargs['df_train_folds']
        df_valid_folds = kwargs['df_valid_folds']
        df_test_folds = kwargs['df_test_folds']

        # self.dt_info = kwargs['dt_info']
        # self.dt_info['n_pcount_list'] = []
        # self.dt_info['n_rgap_list'] = []
        # self.dt_info['n_sgap_list'] = []

        kwargs['cluster_list'] = []

        for idx, (train_dict, test_dict) in enumerate(zip(df_train_folds, df_test_folds)):
            self.train_dict = train_dict
            self.test_dict = test_dict
            if df_valid_folds is not None and len(df_valid_folds) > 0:
                self.val_dict = df_valid_folds[idx]

            self.add_seg_num_features()
            self.cluster = self.construct_cluster_feats()
            kwargs['cluster_list'].append(self.cluster)

        return kwargs

    def add_seg_num_features(self):
        self.train_dict['seg_seq:token_seq'] = self.add_seg_num(self.train_dict)
        self.test_dict['seg_seq:token_seq'] = self.add_seg_num(self.test_dict)
        if self.val_dict is not None:
            self.val_dict['seg_seq:token_seq'] = self.add_seg_num(self.val_dict)

    def add_seg_num(self, data_dict):
        stu_id_in_dict = np.unique(data_dict['stu_id:token'])
        seg_seq = torch.zeros_like(torch.from_numpy(data_dict['exer_seq:token_seq']))
        segs = torch.from_numpy(np.arange(self.window_size) // self.seg_length)

        for s in stu_id_in_dict:
            mask_s = (data_dict['stu_id:token'] == s)
            seg_id_s = torch.arange(0, mask_s.sum())

            seg_seq[mask_s] = segs.repeat((len(seg_id_s),1)) + seg_id_s.unsqueeze(dim=1).repeat((1,self.window_size)) * (self.window_size // self.seg_length)

            # seg_seq.append(seg_id_s)
        return seg_seq

    def construct_cluster_feats(self):

        self.centroids_update_epoch = self.default_cfg['centroids_update_epoch']
        # self.identifiers = self.default_cfg['identifiers']


        cluster_data = {}
        cluster_data['train_cluster_data'] = self._construct_cluster_matrix(self.train_dict)
        if self.val_dict:
            cluster_data['val_cluster_data'] = self._construct_cluster_matrix(self.val_dict)
        cluster_data['test_cluster_data'] = self._construct_cluster_matrix(self.test_dict)
        cluster = self._k_means_cluster(cluster_data)
        return cluster

    def _construct_cluster_matrix(self, data_dict):
        cluster_data = []
        xtotal = torch.zeros((self.stu_count, self.exer_count))
        x1 = torch.zeros((self.stu_count, self.exer_count))
        x0 = torch.zeros((self.stu_count, self.exer_count))

        for i, s in enumerate(data_dict['stu_id:token']):
            exer_s = torch.from_numpy(data_dict['exer_seq:token_seq'][i])
            label_s = torch.from_numpy(data_dict['label_seq:float_seq'][i])
            seg_id_s =  data_dict['seg_seq:token_seq'][i]

            for si in range(len(seg_id_s.unique())):
                stu_id = s
                mask_seg = (seg_id_s == si)
                # seg_id = seg_id_s[mask_seg]
                problem_ids = exer_s[mask_seg]
                correctness = label_s[mask_seg]

                xtotal[stu_id, problem_ids] += 1
                x1[stu_id, problem_ids] += (correctness == 1).float()
                x0[stu_id, problem_ids] += (correctness != 1).float()

                xsr = x1[stu_id] / (xtotal[stu_id] + 0.0000001)
                xfr = x0[stu_id] / (xtotal[stu_id] + 0.0000001)

                x = torch.nan_to_num(xsr) - torch.nan_to_num(xfr)
                x = torch.cat((x, torch.tensor([stu_id])))
                x = torch.cat((x, torch.tensor([si])))
                cluster_data.append(x.cpu().numpy())
        return cluster_data

    def _euclidean_distance(self, x, y):
        return torch.dist(torch.tensor(x), y, p=2)

    def _k_means_cluster(self, cluster_data):
        cluster = {}
        points = np.array(cluster_data['train_cluster_data'])[:,:-2]

        centroids = torch.from_numpy(np.random.permutation(points)[:self.num_cluster, :])
        points_e = torch.from_numpy(np.expand_dims(points, axis=0))
        centroids_e = torch.from_numpy(np.expand_dims(centroids, axis=1))
        distances = torch.sum((points_e - centroids_e) ** 2, dim=-1)
        indices = torch.argmin(distances, dim=0)
        clusters = [points[np.where(indices == i)[0]] for i in range(self.num_cluster)]
        new_centroids = torch.cat(
            [torch.mean(torch.from_numpy(clusters[i]), dim=0, keepdim=True) for i in range(self.num_cluster)], dim=0)

        # update centroids
        centroids = new_centroids

        for j in range(self.centroids_update_epoch):
            points_e = torch.from_numpy(np.expand_dims(points, axis=0))
            centroids_e = torch.from_numpy(np.expand_dims(centroids, axis=1))
            distances = torch.sum((points_e - centroids_e) ** 2, dim=-1)
            indices = torch.argmin(distances, dim=0)
            clusters = [points[np.where(indices == i)[0]] for i in range(self.num_cluster)]
            new_centroids = torch.cat(
                [torch.mean(torch.from_numpy(clusters[i]), dim=0, keepdim=True) for i in range(self.num_cluster)],
                dim=0)
            # update centroids
            centroids = new_centroids.numpy()

        # cluster for training data and testing data
        for students in cluster_data.values():
            for i in students:
                inst = i[:-2]
                min_dist = float('inf')
                closest_clust = None
                for j in range(self.num_cluster):
                    cur_dist = np.linalg.norm(inst - centroids[j], ord=2)
                    if cur_dist < min_dist:
                        min_dist = cur_dist
                        closest_clust = j
                # cluster.append([[int(i[-2]), int(i[-1])], closest_clust])
                cluster[int(i[-2]), int(i[-1])] = closest_clust
        return cluster
        # return pd.DataFrame(cluster, columns=['stu_seg_id','cluster_id'])




from .CDInterDataTPL import CDInterDataTPL
from .CDInterExtendsQDataTPL import CDInterExtendsQDataTPL
from .IRRDataTPL import IRRDataTPL
from .HierCDFDataTPL import HierCDFDataTPL
from .CNCDFDataTPL import CNCDFDataTPL
from .MGCDDataTPL import MGCDDataTPL
from .CNCDQDataTPL import CNCDQDataTPL
from .RCDDataTPL import RCDDataTPL
from .CDGKDataTPL import CDGKDataTPL
from .ECDDataTPL import ECDDataTPL
from .DCDDataTPL import DCDDataTPL
from .FAIRDataTPL import FAIRDataTPL



r"""
HierCDF
##########################################

Reference:
    Jiatong Li et al. "HierCDF: A Bayesian Network-Based Hierarchical Cognitive Diagnosis Framework" in KDD 2022.

Reference Code:
    https://github.com/CSLiJT/HCD-code

"""


import torch
import torch.nn as nn
import networkx as nx
from ..gd_basemodel import GDBaseModel
import numpy as np 
import pandas as pd 
from ..utils.components import PosLinear
import torch.nn.functional as F

def irt2pl(user_emb: torch.Tensor, item_emb: torch.Tensor, item_offset: torch.Tensor):
    return 1 / (1 + torch.exp(-1.7*item_offset*(user_emb - item_emb) ))

def mirt2pl(user_emb: torch.Tensor, item_emb: torch.Tensor, item_offset: torch.Tensor):
    return 1 / (1 + torch.exp(- torch.sum(torch.mul(user_emb, item_emb), axis=1).reshape(-1,1) + item_offset))

def sigmoid_dot(user_emb: torch.Tensor, item_emb: torch.Tensor, item_offset: torch.Tensor):
    return torch.sigmoid(torch.sum(torch.mul(user_emb, item_emb), axis = -1)).reshape(-1,1)

def dot(user_emb: torch.Tensor, item_emb: torch.Tensor, item_offset: torch.Tensor):
    return torch.sum(torch.mul(user_emb, item_emb), axis = -1).reshape(-1,1)

itf_dict = {
    'irt': irt2pl,
    'mirt': mirt2pl,
    'mf': dot, 
    'sigmoid-mf': sigmoid_dot
}

class HierCDF(GDBaseModel):
    default_cfg = {
        'itf_type': 'mirt',
        'hidden_dim': 1,
        'lambda': 0.001, 
    }
    def __init__(self, cfg):
        super().__init__(cfg)

    def build_cfg(self):
        self.n_user = self.datatpl_cfg['dt_info']['stu_count']
        self.n_item = self.datatpl_cfg['dt_info']['exer_count']
        self.n_cpt = self.datatpl_cfg['dt_info']['cpt_count']
        self.hidden_dim = self.modeltpl_cfg['hidden_dim']
        self.itf_type = self.modeltpl_cfg['itf_type']

        self.set_itf(self.itf_type)

    def build_model(self):
        # the conditional mastery degree when parent is mastered
        self.condi_p = nn.Embedding(self.n_user, self.know_graph.shape[0])

        # the conditional mastery degree when parent is non-mastered
        self.condi_n = nn.Embedding(self.n_user, self.know_graph.shape[0])

        # the priori mastery degree of parent
        self.priori = nn.Embedding(self.n_user, self.n_cpt)

        # item representation
        self.item_diff = nn.Embedding(self.n_item, self.n_cpt)
        self.item_disc = nn.Embedding(self.n_item, 1)

        # embedding transformation
        self.user_contract = PosLinear(self.n_cpt, self.hidden_dim)
        self.item_contract = PosLinear(self.n_cpt, self.hidden_dim)

        # Neural Interaction Module (used only in ncd)
        self.cross_layer1 = PosLinear(self.hidden_dim,max(int(self.hidden_dim/2),1))
        self.cross_layer2 = PosLinear(max(int(self.hidden_dim/2),1),1)

    def add_extra_data(self, df_cpt_relation, **kwargs):
        self.know_graph = df_cpt_relation
        self.know_edge = nx.DiGraph()  #nx.DiGraph(know_graph.values.tolist())
        for k in range(self.n_cpt):
            self.know_edge.add_node(k)
        for edge in df_cpt_relation[['cpt_head', 'cpt_tail']].to_numpy():
            self.know_edge.add_edge(edge[0],edge[1])
        self.topo_order = list(nx.topological_sort(self.know_edge))
        self.Q_mat = kwargs['Q_mat'].to(self.device)

    def ncd(self, user_emb: torch.Tensor, item_emb: torch.Tensor, item_offset: torch.Tensor):
        input_vec = (user_emb-item_emb)*item_offset
        x_vec=torch.sigmoid(self.cross_layer1(input_vec))
        x_vec=torch.sigmoid(self.cross_layer2(x_vec))
        return x_vec
    
    def set_itf(self, itf_type):
        self.itf_type = itf_type
        self.itf = itf_dict.get(itf_type, self.ncd)

    def get_posterior(self, user_ids: torch.LongTensor):
        n_batch = user_ids.shape[0]
        posterior = torch.rand(n_batch, self.n_cpt).to(self.device)
        batch_priori = torch.sigmoid(self.priori(user_ids))
        batch_condi_p = torch.sigmoid(self.condi_p(user_ids))
        batch_condi_n = torch.sigmoid(self.condi_n(user_ids))
        
        for k in self.topo_order:
            # get predecessor list
            predecessors = list(self.know_edge.predecessors(k))
            predecessors.sort()
            len_p = len(predecessors)

            # for each knowledge k, do:
            if len_p == 0:
                priori = batch_priori[:,k]
                posterior[:,k] = priori.reshape(-1)
                continue

            # format of masks
            tpl = '{0:0%db}'%(len_p)
            # number of parent master condition
            n_condi = 2 ** len_p

            priori = posterior[:,predecessors]


            pred_idx = self.know_graph[self.know_graph['cpt_tail'] == k].sort_values(by='cpt_head').index
            condi_p = torch.pow(batch_condi_p[:,pred_idx],1/len_p)
            condi_n = torch.pow(batch_condi_n[:,pred_idx],1/len_p)
            
            margin_p = condi_p * priori
            margin_n = condi_n * (1.0-priori)

            posterior_k = torch.zeros((1,n_batch)).to(self.device)

            for idx in range(n_condi):
                # for each parent mastery condition, do:
                mask = tpl.format(idx)
                mask = torch.Tensor(np.array(list(mask)).astype(int)).to(self.device)

                margin = mask * margin_p + (1-mask) * margin_n
                margin = torch.prod(margin, dim = 1).unsqueeze(dim = 0)

                posterior_k = torch.cat([posterior_k, margin], dim = 0)
            posterior_k = (torch.sum(posterior_k, dim = 0)).squeeze()
            
            posterior[:,k] = posterior_k.reshape(-1)
        return posterior
    
    def get_condi_p(self,user_ids: torch.LongTensor, device = 'cpu')->torch.Tensor:
        n_batch = user_ids.shape[0]
        result_tensor = torch.rand(n_batch, self.n_cpt).to(device)
        batch_priori = torch.sigmoid(self.priori(user_ids))
        batch_condi_p = torch.sigmoid(self.condi_p(user_ids))
        
        for k in self.topo_order:
            # get predecessor list
            predecessors = list(self.know_edge.predecessors(k))
            predecessors.sort()
            len_p = len(predecessors)
            if len_p == 0:
                priori = batch_priori[:,k]
                result_tensor[:,k] = priori.reshape(-1)
                continue
            pred_idx = self.know_graph[self.know_graph['cpt_tail'] == k].sort_values(by='cpt_head').index
            condi_p = torch.pow(batch_condi_p[:,pred_idx],1/len_p)
            result_tensor[:,k] = torch.prod(condi_p, dim=1).reshape(-1)
        
        return result_tensor

    def get_condi_n(self,user_ids: torch.LongTensor, device = 'cpu')->torch.Tensor:
        n_batch = user_ids.shape[0]
        result_tensor = torch.rand(n_batch, self.n_cpt).to(device)
        batch_priori = torch.sigmoid(self.priori(user_ids))
        batch_condi_n = torch.sigmoid(self.condi_n(user_ids))
        
        for k in self.topo_order:
            # get predecessor list
            predecessors = list(self.know_edge.predecessors(k))
            predecessors.sort()
            len_p = len(predecessors)
            if len_p == 0:
                priori = batch_priori[:,k]
                result_tensor[:,k] = priori.reshape(-1)
                continue
            pred_idx = self.know_graph[self.know_graph['cpt_tail'] == k].sort_values(by='cpt_head').index
            condi_n = torch.pow(batch_condi_n[:,pred_idx],1/len_p)
            result_tensor[:,k] = torch.prod(condi_n, dim=1).reshape(-1)
        
        return result_tensor

    def forward(self, stu_id: torch.LongTensor, exer_id: torch.LongTensor):
        Q_mat = self.Q_mat[exer_id]
        user_mastery = self.get_posterior(stu_id)
        item_diff = torch.sigmoid(self.item_diff(exer_id))
        item_disc = torch.sigmoid(self.item_disc(exer_id))

        user_factor = torch.tanh(self.user_contract(user_mastery * Q_mat))
        item_factor = torch.sigmoid(self.item_contract(item_diff * Q_mat))
        
        output = self.itf(user_factor, item_factor, item_disc)

        return output


    @torch.no_grad()
    def predict(self, stu_id, exer_id, **kwargs):
        return {
            'y_pd': self(stu_id, exer_id).flatten(),
        }

    def get_main_loss(self, **kwargs):
        stu_id = kwargs['stu_id']
        exer_id = kwargs['exer_id']
        label = kwargs['label']
        pd = self(stu_id, exer_id).flatten()
        loss = F.binary_cross_entropy(input=pd, target=label)
        return {
            'loss_main': loss
        }

    def get_J_loss(self, **kwargs):
        stu_id = kwargs['stu_id']
        return {
            'loss_J': self.modeltpl_cfg['lambda'] * \
                torch.sum(torch.relu(self.condi_n(stu_id)-self.condi_p(stu_id)))
        }
    
    def get_loss_dict(self, **kwargs):
        ret_dict = self.get_main_loss(**kwargs)
        ret_dict.update(self.get_J_loss(**kwargs))
        return ret_dict
    
    def get_stu_status(self, stu_id=None):
        if stu_id is not None:
            return self.get_posterior(stu_id)
        else:
            stu_id = torch.arange(self.n_user).to(self.device)
            return self.get_posterior(stu_id)



# reference: https://github.com/keras-team/keras/blob/v2.12.0/keras/utils/data_utils.py#L1009-L1143

import numpy as np
from collections import defaultdict
import logging


logger = logging.getLogger("edustudio")


class PadSeqUtil(object):
    @staticmethod
    def pad_sequence(
            sequences, maxlen=0, dtype='int64', padding='post',
            is_truncate=False, truncating='post', value=0.0, 
            return_idx=False, return_mask=False,
        ):
        if not hasattr(sequences, "__len__"):
            raise ValueError("`sequences` must be iterable.")
        num_samples = len(sequences)

        lengths = []
        sample_shape = ()
        flag = True

        # take the sample shape from the first non empty sequence
        # checking for consistency in the main loop below.

        for x in sequences:
            try:
                lengths.append(len(x))
                if flag and len(x):
                    sample_shape = np.asarray(x).shape[1:]
                    flag = False
            except TypeError as e:
                raise ValueError(
                    "`sequences` must be a list of iterables. "
                    f"Found non-iterable: {str(x)}"
                ) from e

        if maxlen <= 0 or maxlen is None:
            maxlen = np.max(lengths)


        return_idx_list = []
        if not is_truncate:
            num_samples = 0
            for idx, len_ in enumerate(lengths):
                num = int(np.ceil(len_ / maxlen))
                num_samples += num
                if return_idx:
                    return_idx_list.append([idx] * num)

        if return_idx:
            return_idx = np.concatenate(return_idx_list).astype(np.int64)
        
        version = np.__version__

        if version.startswith('2.'):
            is_dtype_str = np.issubdtype(dtype, np.str_)
        else:
            is_dtype_str = np.issubdtype(dtype, np.str_) or np.issubdtype(
                dtype, np.unicode_
            )
        if isinstance(value, str) and dtype != object and not is_dtype_str:
            raise ValueError(
                f"`dtype` {dtype} is not compatible with `value`'s type: "
                f"{type(value)}\nYou should set `dtype=object` for variable length "
                "strings."
            )

        x = np.full((num_samples, maxlen) + sample_shape, value, dtype=dtype)
        mask = None
        if return_mask:
            mask = np.full((num_samples, maxlen) + sample_shape, 1, dtype=np.int8)
        idx = 0
        flag = True
        for s in sequences:
            if not len(s):
                if flag:
                    logger.warning("empty list/array was found")
                    flag = False
                continue

            if is_truncate:
                if truncating == "pre":
                    trunc = s[-maxlen:]
                elif truncating == "post":
                    trunc = s[:maxlen]
                else:
                    raise ValueError(f'Truncating type "{truncating}" not understood')

                # check `trunc` has expected shape
                trunc = np.asarray(trunc, dtype=dtype)
                if trunc.shape[1:] != sample_shape:
                    raise ValueError(
                        f"Shape of sample {trunc.shape[1:]} of sequence at "
                        f"position {idx} is different from expected shape "
                        f"{sample_shape}"
                    )

                if padding == "post":
                    x[idx, : len(trunc)] = trunc
                    if return_mask:
                        mask[idx, len(trunc):] = 0
                elif padding == "pre":
                    x[idx, -len(trunc) :] = trunc
                    if return_mask:
                        mask[idx, :-len(trunc)] = 0
                else:
                    raise ValueError(f'Padding type "{padding}" not understood')
                idx += 1
            else:
                for i in range(int(np.ceil(len(s) / maxlen))):
                    trunc = s[i*maxlen:maxlen*(i+1)]
                    # check `trunc` has expected shape
                    trunc = np.asarray(trunc, dtype=dtype)
                    if trunc.shape[1:] != sample_shape:
                        raise ValueError(
                            f"Shape of sample {trunc.shape[1:]} of sequence at "
                            f"position {idx} is different from expected shape "
                            f"{sample_shape}"
                        )
                    if padding == "post":
                        x[idx, : len(trunc)] = trunc
                        if return_mask:
                            mask[idx, len(trunc):] = 0
                    elif padding == "pre":
                        x[idx, -len(trunc) :] = trunc
                        if return_mask:
                            mask[idx, :-len(trunc)] = 0
                    else:
                        raise ValueError(f'Padding type "{padding}" not understood')
                    idx += 1
        return x, return_idx, mask


if __name__ == '__main__':
    a = PadSeqUtil.pad_sequence(
        [[1],[2,1,3,4,4,4,5,0,5,19,6],[],[0,1,1,1,1,1]], maxlen=4, is_truncate=False, value=-1, return_idx=True, return_mask=True, padding='pre')
    print(a)



from ..common.base_mid2cache import BaseMid2Cache
import pandas as pd
import numpy as np
from edustudio.datatpl.utils import SpliterUtil, PadSeqUtil
from itertools import chain


class M2C_RandomDataSplit4KT(BaseMid2Cache):
    default_cfg = {
        'seed': 2023,
        'divide_by': 'stu',
        "divide_scale_list": [7,1,2],
    }

    def __init__(self, m2c_cfg, n_folds, is_dataset_divided) -> None:
        super().__init__(m2c_cfg)
        self.n_folds = n_folds
        self.is_dataset_divided = is_dataset_divided

    @classmethod
    def from_cfg(cls, cfg):
        m2c_cfg = cfg.datatpl_cfg.get(cls.__name__)
        n_folds = cfg.datatpl_cfg.n_folds
        is_dataset_divided = cfg.datatpl_cfg.is_dataset_divided
        return cls(m2c_cfg, n_folds, is_dataset_divided)

    def _check_params(self):
        super()._check_params()
        assert self.m2c_cfg['divide_by'] in {'stu', 'time'}

    def process(self, **kwargs):
        df_seq = kwargs['df_seq']
        df_train_seq = kwargs.get('df_train_seq', None)
        df_valid_seq = kwargs.get('df_validn_seq', None)
        df_test_seq = kwargs.get('df_test_seq', None)

        if not self.is_dataset_divided:
            assert df_train_seq is None and df_valid_seq is None and df_test_seq is None
            self.window_size = df_seq['exer_seq:token_seq'].shape[1]
            if self.m2c_cfg['divide_by'] == 'stu':
                if self.n_folds == 1:
                    train_dict, valid_dict, test_dict = self._divide_data_df_by_stu_one_fold(df_seq)
                    kwargs['df_train_folds'] = [train_dict]
                    kwargs['df_valid_folds'] = [valid_dict]
                    kwargs['df_test_folds'] = [test_dict]
                else:
                    kwargs['df_train_folds'], kwargs['df_valid_folds'], kwargs['df_test_folds'] = self._divide_data_df_by_stu_multi_fold(df_seq)
            elif self.m2c_cfg['divide_by'] == 'time':
                raise NotImplementedError
            else:
                raise ValueError(f"unknown divide_by: {self.m2c_cfg['divide_by']}")
        else:
            assert df_train_seq is not None and df_test_seq is not None
            self.window_size = df_train_seq['exer_seq:token_seq'].shape[1]
            kwargs['df_train_folds'] = [df_train_seq]
            kwargs['df_valid_folds'] = [df_valid_seq]
            kwargs['df_test_folds'] = [df_test_seq]
        return kwargs

    def _dict_index_flag(self, df_seq:dict, flag: np.array):
        return {
            k: df_seq[k][flag] for k in df_seq
        }

    def _divide_data_df_by_stu_one_fold(self, df_seq: dict):
        train_stu_id, valid_stu_id, test_stu_id = SpliterUtil.divide_data_df_one_fold(
            pd.DataFrame({"stu_id:token": np.unique(df_seq['stu_id:token'])}), seed=self.m2c_cfg['seed'], shuffle=True,
            divide_scale_list=self.m2c_cfg['divide_scale_list']
        )

        df_train_seq = self._dict_index_flag(df_seq, np.isin(df_seq['stu_id:token'], train_stu_id.to_numpy().flatten()))
        df_test_seq = self._dict_index_flag(df_seq, np.isin(df_seq['stu_id:token'], test_stu_id.to_numpy().flatten()))
        df_valid_seq = None
        if valid_stu_id is not None:
            df_valid_seq = self._dict_index_flag(df_seq, np.isin(df_seq['stu_id:token'], valid_stu_id.to_numpy().flatten()))

        return df_train_seq, df_test_seq, df_valid_seq
    
    def _divide_data_df_by_stu_multi_fold(self, df_seq: pd.DataFrame):
        res = SpliterUtil.divide_data_df_multi_folds(
            pd.DataFrame({"stu_id:token": np.unique(df_seq['stu_id:token'])}), seed=self.m2c_cfg['seed'], shuffle=True, n_folds=self.n_folds
        )

        train_list,  test_list = [], []
        for (train_stu_id, test_stu_id) in zip(*res):
            df_train_seq = self._dict_index_flag(df_seq, np.isin(df_seq['stu_id:token'], train_stu_id.to_numpy().flatten()))
            df_test_seq = self._dict_index_flag(df_seq, np.isin(df_seq['stu_id:token'], test_stu_id.to_numpy().flatten()))
            train_list.append(df_train_seq)
            test_list.append(df_test_seq)

        return train_list, [], test_list

    def set_dt_info(self, dt_info, **kwargs):
        dt_info['real_window_size'] = self.window_size
        if not self.is_dataset_divided:
            if 'stu_id:token' in kwargs['df'].columns:
                dt_info['stu_count'] = int(kwargs['df']['stu_id:token'].max() + 1)
            if 'exer_id:token' in kwargs['df'].columns:
                dt_info['exer_count'] = int(kwargs['df']['exer_id:token'].max() + 1)
        else:
            stu_count = max(kwargs['df_train']['stu_id:token'].max() + 1, kwargs['df_test']['stu_id:token'].max() + 1)
            stu_count = max(kwargs['df_valid']['stu_id:token'].max() + 1, stu_count) if 'df_valid' in kwargs else stu_count

            exer_count = max(kwargs['df_train']['exer_id:token'].max() + 1, kwargs['df_test']['exer_id:token'].max() + 1)
            exer_count = max(kwargs['df_valid']['exer_id:token'].max() + 1, exer_count) if 'df_valid' in kwargs else exer_count

            dt_info['stu_count'] = stu_count
            dt_info['exer_count'] = exer_count

        if kwargs.get('df_exer', None) is not None:
            if 'cpt_seq:token_seq' in kwargs['df_exer']:
                dt_info['cpt_count'] = len(set(list(chain(*kwargs['df_exer']['cpt_seq:token_seq'].to_list()))))



import torch.nn as nn
import torch
from abc import abstractmethod
from .utils.common import xavier_normal_initialization, xavier_uniform_initialization, kaiming_normal_initialization, kaiming_uniform_initialization
from .basemodel import BaseModel


class GDBaseModel(BaseModel):
    """
    The model that using gradient descent method can inherit the class
    """
    default_cfg = {
        'param_init_type': 'xavier_normal', # initialization method of model paramters
        'pretrained_file_path': "", # file path of pretrained model parameters
    }

    def __init__(self, cfg):
        super().__init__(cfg)
        self.device = self.traintpl_cfg['device']
        self.share_callback_dict = {
            "stop_training": False
        }

    @abstractmethod
    def build_cfg(self):
        """Construct model config
        """
        pass

    @abstractmethod
    def build_model(self):
        """Construct model component
        """
        pass

    def _init_params(self):
        """Initialize the model parameters
        """
        if self.modeltpl_cfg['param_init_type'] == 'default':
            pass
        elif self.modeltpl_cfg['param_init_type'] == 'xavier_normal':
            self.apply(xavier_normal_initialization)
        elif self.modeltpl_cfg['param_init_type'] == 'xavier_uniform':
            self.apply(xavier_uniform_initialization)
        elif self.modeltpl_cfg['param_init_type'] == 'kaiming_normal':
            self.apply(kaiming_normal_initialization)
        elif self.modeltpl_cfg['param_init_type'] == 'kaiming_uniform':
            self.apply(kaiming_uniform_initialization)
        elif self.modeltpl_cfg['param_init_type'] == 'init_from_pretrained':
            self._load_params_from_pretrained()

    def _load_params_from_pretrained(self):
        """Load pretrained model parameters
        """
        self.load_state_dict(torch.load(self.modeltpl_cfg['pretrained_file_path']))

    def predict(self, **kwargs):
        """predict process
        """
        pass

    def get_loss_dict(self, **kwargs):
        """Get a dict object. The key is the loss name, the value is the loss
        """
        pass



r"""
IEKT
##########################################

Reference:
    Ting Long et al. "Tracing Knowledge State with Individual Cognition and Acquisition Estimation" in SIGIR 2021.

Reference Code:
    https://github.com/ApexEDM/iekt

"""

from ..gd_basemodel import GDBaseModel
import torch.nn as nn
import torch
import torch.nn.functional as F
from ..utils.components import MLP
from torch.distributions import Categorical


class mygru(nn.Module):
    '''
    classifier decoder implemented with mlp
    '''
    def __init__(self, n_layer, input_dim, hidden_dim):
        super().__init__()
        
        this_layer = n_layer
        self.g_ir = funcsgru(this_layer, input_dim, hidden_dim, 0)
        self.g_iz = funcsgru(this_layer, input_dim, hidden_dim, 0)
        self.g_in = funcsgru(this_layer, input_dim, hidden_dim, 0)
        self.g_hr = funcsgru(this_layer, hidden_dim, hidden_dim, 0)
        self.g_hz = funcsgru(this_layer, hidden_dim, hidden_dim, 0)
        self.g_hn = funcsgru(this_layer, hidden_dim, hidden_dim, 0)
        self.sigmoid = torch.nn.Sigmoid()
        self.tanh = torch.nn.Tanh()

    def forward(self, x, h):
        r_t = self.sigmoid(
            self.g_ir(x) + self.g_hr(h)
        )
        z_t = self.sigmoid(
            self.g_iz(x) + self.g_hz(h)
        )
        n_t = self.tanh(
            self.g_in(x) + self.g_hn(h).mul(r_t)
        )
        h_t = (1 - z_t) * n_t + z_t * h
        return h_t
    
class funcsgru(nn.Module):
    '''
    classifier decoder implemented with mlp
    '''
    def __init__(self, n_layer, hidden_dim, output_dim, dpo):
        super().__init__()

        self.lins = nn.ModuleList([
            nn.Linear(hidden_dim, hidden_dim)
            for _ in range(n_layer)
        ])
        self.dropout = nn.Dropout(p = dpo)
        self.out = nn.Linear(hidden_dim, output_dim)

    def forward(self, x):
        for lin in self.lins:
            x = F.relu(lin(x))
        return self.out(self.dropout(x))


class IEKT(GDBaseModel):
    default_cfg = {
        'd_q': 64, # embedding size of exercise embedding
        'd_c': 64, # embedding size of concept embedding
        'd_m': 128, # embedding size of cognition levels matrix
        'd_s': 128, # embedding size of acquisition sensitivity levels matrix
        'd_h': 64, # size of hidden state
        'n_cog_level': 10, 
        'n_acq_level': 10,
        'dropout_rate': 0.0,
        'gamma': 0.93,
        'lambda': 40.0,
    }

    def build_cfg(self):
        self.n_stu = self.datatpl_cfg['dt_info']['stu_count']
        self.n_exer = self.datatpl_cfg['dt_info']['exer_count']
        self.n_cpt = self.datatpl_cfg['dt_info']['cpt_count']
        self.window_size = self.datatpl_cfg['dt_info']['real_window_size']
        self.n_cog_level = self.modeltpl_cfg['n_cog_level']
        self.n_acq_level = self.modeltpl_cfg['n_acq_level']
        self.d_v = self.modeltpl_cfg['d_h'] + self.modeltpl_cfg['d_q'] + self.modeltpl_cfg['d_c']
        self.d_r = self.d_v + self.modeltpl_cfg['d_m']
        self.d_i = self.modeltpl_cfg['d_q'] + self.modeltpl_cfg['d_c'] + self.modeltpl_cfg['d_s']

    def build_model(self):
        self.exer_emb = nn.Embedding(self.n_exer, self.modeltpl_cfg['d_q'])
        self.cpt_emb = nn.Embedding(self.n_cpt, self.modeltpl_cfg['d_c'])
        self.cog_matrix = nn.Embedding(self.n_cog_level, self.modeltpl_cfg['d_m'])
        self.acq_matrix = nn.Embedding(self.n_acq_level, self.modeltpl_cfg['d_s'])

        self.pd_layer = MLP(
            input_dim=self.d_r, output_dim=1, 
            dnn_units=[self.d_r], dropout_rate=self.modeltpl_cfg['dropout_rate']
        )
        self.f_p = MLP(
            input_dim=self.d_v, output_dim=self.n_cog_level, 
            dnn_units=[self.d_r], dropout_rate=self.modeltpl_cfg['dropout_rate']
        )
        self.f_e = MLP(
            input_dim=self.d_v * 4, output_dim=self.n_acq_level,
            dnn_units=[self.d_r], dropout_rate=self.modeltpl_cfg['dropout_rate']
        )
        self.gru_h = mygru(0, self.d_i, self.modeltpl_cfg['d_h'])

    def get_exer_representation(self, exer_ids, cpt_seq, cpt_seq_mask):
        exer_emb = self.exer_emb(exer_ids) # windows_size x emb size
        cpt_emb = torch.vstack(
            [self.cpt_emb(cpt_seq[i])[cpt_seq_mask[i] == 1].mean(dim=0) for i in range(cpt_seq.shape[0])]
        )
        return torch.cat([exer_emb, cpt_emb], dim = -1)

    def pi_cog_func(self, x, softmax_dim = 1):
        return F.softmax(self.f_p(x), dim = softmax_dim)
    
    def pi_sens_func(self, x, softmax_dim = 1):
        return F.softmax(self.f_e(x), dim = softmax_dim)
    
    def update_state(self, h, v, s_t, operate):
        v_cat = torch.cat([
            v.mul(operate.reshape(-1,1)),
            v.mul((1 - operate).reshape(-1,1))], dim = 1)
        e_cat = torch.cat([
            s_t.mul((1-operate).reshape(-1,1)),
            s_t.mul((operate).reshape(-1,1))], dim = 1)
        inputs = v_cat + e_cat
        next_p_state = self.gru_h(inputs, h)
        return next_p_state
     
    def get_main_loss(self, **kwargs):
        exer_seq = kwargs['exer_seq']
        exer_mask_seq = kwargs['mask_seq']
        cpt_seq = kwargs['cpt_seq']
        cpt_seq_mask = kwargs['cpt_seq_mask']
        label_seq = kwargs['label_seq']

        batch_size = exer_seq.shape[0]
        seq_len = exer_seq.shape[1]
        h = torch.zeros(batch_size, self.modeltpl_cfg['d_h']).to(self.device)
        p_action_list, pre_state_list, emb_action_list, states_list, reward_list, predict_list, ground_truth_list = [], [], [], [], [], [], []


        for t in range(seq_len):
            # read stage
            exer_seq_col = exer_seq[:, t]
            cpt_seq_col = cpt_seq[:, t]
            cpt_seq_mask_col = cpt_seq_mask[:, t]
            label_seq_col = label_seq[:, t]
            v = self.get_exer_representation(exer_seq_col,  cpt_seq_col, cpt_seq_mask_col)
            h_v = torch.cat([h, v], dim=1)
            v_h = torch.cat([v, h], dim=1)
            flip_prob_emb = self.pi_cog_func(v_h)

            m = Categorical(flip_prob_emb) # 从多项式分布中sample
            cog_sample = m.sample()
            m_t = self.cog_matrix(cog_sample) # 从CE矩阵取出对应emb
            
            logits = self.pd_layer(torch.cat([h_v, m_t], dim = 1)) # 得到预测结果
            prob = logits.sigmoid()

            # write stage
            v_g = torch.cat([
                h_v.mul(label_seq_col.reshape(-1,1).float()),
                h_v.mul((1-label_seq_col).reshape(-1,1).float())],
                dim = 1)

            out_operate_logits = torch.where(prob > 0.5, torch.tensor(1).to(self.device), torch.tensor(0).to(self.device))  # 将预测值离散化
            v_p = torch.cat([
                h_v.mul(out_operate_logits.reshape(-1,1).float()),
                h_v.mul((1-out_operate_logits).reshape(-1,1).float())],
                dim = 1)                
            v_m = torch.cat([v_g, v_p], dim = 1)

            flip_prob_emb = self.pi_sens_func(v_m)
            m = Categorical(flip_prob_emb)
            acq_sample = m.sample()
            s_t = self.acq_matrix(acq_sample)
            h = self.update_state(h, v, s_t, label_seq_col)

            emb_action_list.append(cog_sample)
            p_action_list.append(acq_sample)
            states_list.append(v_m)
            pre_state_list.append(v_h)
            
            ground_truth_list.append(label_seq_col)
            predict_list.append(logits.squeeze(1))
            this_reward = torch.where(out_operate_logits.squeeze(1).float() == label_seq_col,
                            torch.tensor(1).to(self.device), 
                            torch.tensor(0).to(self.device))
            reward_list.append(this_reward)

        # RL 训练
        seq_num = exer_mask_seq.sum(dim=1)
        emb_action_tensor = torch.stack(emb_action_list, dim = 1) # ac的sample操作
        p_action_tensor = torch.stack(p_action_list, dim = 1) # ce的sample操作
        state_tensor = torch.stack(states_list, dim = 1) # 输出的groudtruth和predict合并
        pre_state_tensor = torch.stack(pre_state_list, dim = 1)
        reward_tensor = torch.stack(reward_list, dim = 1).float() / (seq_num.unsqueeze(-1).repeat(1, self.window_size)).float()
        logits_tensor = torch.stack(predict_list, dim = 1)
        ground_truth_tensor = torch.stack(ground_truth_list, dim = 1)
        loss = []
        tracat_logits = []
        tracat_ground_truth = []
        
        for i in range(0, batch_size):
            this_seq_len = seq_num[i]
            this_reward_list = reward_tensor[i]
        
            this_cog_state = torch.cat([pre_state_tensor[i][0: this_seq_len],
                                    torch.zeros(1, pre_state_tensor[i][0].size()[0]).to(self.device)
                                    ], dim = 0)
            this_sens_state = torch.cat([state_tensor[i][0: this_seq_len],
                                    torch.zeros(1, state_tensor[i][0].size()[0]).to(self.device)
                                    ], dim = 0)

            td_target_cog = this_reward_list[0: this_seq_len].unsqueeze(1)
            delta_cog = td_target_cog
            delta_cog = delta_cog.detach().cpu().numpy()

            td_target_sens = this_reward_list[0: this_seq_len].unsqueeze(1)
            delta_sens = td_target_sens
            delta_sens = delta_sens.detach().cpu().numpy()

            advantage_lst_cog = []
            advantage = 0.0
            for delta_t in delta_cog[::-1]:
                advantage = self.modeltpl_cfg['gamma'] * advantage + delta_t[0]
                advantage_lst_cog.append([advantage])
            advantage_lst_cog.reverse()
            advantage_cog = torch.tensor(advantage_lst_cog, dtype=torch.float).to(self.device)
            
            pi_cog = self.pi_cog_func(this_cog_state[:-1])
            pi_a_cog = pi_cog.gather(1,p_action_tensor[i][0: this_seq_len].unsqueeze(1))

            loss_cog = -torch.log(pi_a_cog) * advantage_cog
            
            loss.append(torch.sum(loss_cog))

            advantage_lst_sens = []
            advantage = 0.0
            for delta_t in delta_sens[::-1]:
                advantage = self.modeltpl_cfg['gamma'] * advantage + delta_t[0]
                advantage_lst_sens.append([advantage])
            advantage_lst_sens.reverse()
            advantage_sens = torch.tensor(advantage_lst_sens, dtype=torch.float).to(self.device)
            
            pi_sens = self.pi_sens_func(this_sens_state[:-1])
            pi_a_sens = pi_sens.gather(1,emb_action_tensor[i][0: this_seq_len].unsqueeze(1))

            loss_sens = - torch.log(pi_a_sens) * advantage_sens
            loss.append(torch.sum(loss_sens))
            
            this_prob = logits_tensor[i][0: this_seq_len]
            this_groud_truth = ground_truth_tensor[i][0: this_seq_len]

            tracat_logits.append(this_prob)
            tracat_ground_truth.append(this_groud_truth)
        

        bce = F.binary_cross_entropy_with_logits(torch.cat(tracat_logits, dim = 0), torch.cat(tracat_ground_truth, dim = 0))
           
        label_len = torch.cat(tracat_ground_truth, dim = 0).size()[0]
        loss_l = sum(loss)
        loss = self.modeltpl_cfg['lambda'] * (loss_l / label_len) +  bce
        return {
            'loss_main': loss
        }

    def get_loss_dict(self, **kwargs):
        return self.get_main_loss(**kwargs)


    @torch.no_grad()
    def predict(self, **kwargs):
        exer_seq = kwargs['exer_seq']
        exer_mask_seq = kwargs['mask_seq']
        cpt_seq = kwargs['cpt_seq']
        cpt_seq_mask = kwargs['cpt_seq_mask']
        label_seq = kwargs['label_seq']

        batch_size = exer_seq.shape[0]
        seq_len = exer_seq.shape[1]
        h = torch.zeros(batch_size, self.modeltpl_cfg['d_h']).to(self.device)
        batch_probs, uni_prob_list =[], []


        for t in range(seq_len):
            # read stage
            exer_seq_col = exer_seq[:, t]
            cpt_seq_col = cpt_seq[:, t]
            cpt_seq_mask_col = cpt_seq_mask[:, t]
            label_seq_col = label_seq[:, t]
            v = self.get_exer_representation(exer_seq_col,  cpt_seq_col, cpt_seq_mask_col)
            h_v = torch.cat([h, v], dim=1)
            v_h = torch.cat([v, h], dim=1)
            flip_prob_emb = self.pi_cog_func(v_h)

            m = Categorical(flip_prob_emb) # 从多项式分布中sample
            cog_sample = m.sample()
            m_t = self.cog_matrix(cog_sample) # 从CE矩阵取出对应emb
            
            logits = self.pd_layer(torch.cat([h_v, m_t], dim = 1)) # 得到预测结果
            prob = logits.sigmoid()

            # write stage
            v_g = torch.cat([
                h_v.mul(label_seq_col.reshape(-1,1).float()),
                h_v.mul((1-label_seq_col).reshape(-1,1).float())],
                dim = 1)

            out_operate_logits = torch.where(prob > 0.5, torch.tensor(1).to(self.device), torch.tensor(0).to(self.device))  # 将预测值离散化
            v_p = torch.cat([
                h_v.mul(out_operate_logits.reshape(-1,1).float()),
                h_v.mul((1-out_operate_logits).reshape(-1,1).float())],
                dim = 1)                
            v_m = torch.cat([v_g, v_p], dim = 1)

            flip_prob_emb = self.pi_sens_func(v_m)
            m = Categorical(flip_prob_emb)
            acq_sample = m.sample()
            s_t = self.acq_matrix(acq_sample)
            h = self.update_state(h, v, s_t, label_seq_col)

            uni_prob_list.append(prob.detach())

        seq_num = exer_mask_seq.sum(dim=1)
        prob_tensor = torch.cat(uni_prob_list, dim = 1)
        for i in range(0, batch_size):
            this_seq_len = seq_num[i]
            batch_probs.append(prob_tensor[i][0: this_seq_len])
        y_pd = torch.cat(batch_probs, dim = 0)

        y_gt = kwargs['label_seq']
        y_gt = y_gt[kwargs['mask_seq'] == 1]
        
        return {
            'y_pd': y_pd,
            'y_gt': y_gt
        }



r"""
QDKT
##########################################

Reference:
    Shashank Sonkar et al. "qDKT: Question-Centric Deep Knowledge Tracing" in EDM 2020.

Reference Code:
    https://github.com/pykt-team/pykt-toolkit/blob/main/pykt/models/qdkt.py

"""
from ..gd_basemodel import GDBaseModel
import torch.nn as nn
import torch
import torch.nn.functional as F

class QDKT(GDBaseModel):
    r"""
   QDKT

   default_cfg:
      'emb_size': 100  # dimension of embedding
      'dropout_rate': 0.2      # dropout rate
      'hidden_size': 100        # hidden size of LSTM
        'num_layers': 1         # num layers of LSTM
        'w_reg':0.01            # weight of loss_reg
   """
    default_cfg = {
        'emb_size': 100,
        'hidden_size': 100,
        'num_layers': 1,
        'dropout_rate': 0.2,
        'w_reg':0.01
    }

    def add_extra_data(self, **kwargs):
        self.laplacian_matrix = kwargs['laplacian_matrix']
        self.laplacian = torch.tensor(self.laplacian_matrix).to(self.device).float()
        self.train_dict = kwargs['train_dict']

    def __init__(self, cfg):
        super().__init__(cfg)

    def build_cfg(self):
        self.n_user = self.datatpl_cfg['dt_info']['stu_count']
        self.n_item = self.datatpl_cfg['dt_info']['exer_count']
        self.w_reg = self.modeltpl_cfg['w_reg']

    def _init_params(self):
        super()._init_params()
        for w in self.fasttext_model.words:
            label = int(w[-1])
            index = int(w[:-1])
            fast_weight = self.fasttext_model[w]
            new_index = index+label*index
            sd = self.exer_emb.state_dict()
            sd['weight'][new_index].copy_(torch.tensor(fast_weight))

    def build_model(self):
        self.exer_emb = nn.Embedding(
            self.n_item * 2, self.modeltpl_cfg['emb_size']
        )
        self.seq_model = nn.LSTM(
            self.modeltpl_cfg['emb_size'], self.modeltpl_cfg['hidden_size'],
            self.modeltpl_cfg['num_layers'], batch_first=True
        )

        self.dropout_layer = nn.Dropout(self.modeltpl_cfg['dropout_rate'])
        self.fc_layer = nn.Linear(self.modeltpl_cfg['hidden_size'], self.n_item)

        f = open("tmp.txt", "w")
        for l in self.train_fasttext():
            f.writelines(l)
        f.close()
        import fasttext
        self.fasttext_model = fasttext.train_unsupervised("tmp.txt",  dim = self.modeltpl_cfg['emb_size'],minCount=1, wordNgrams=2)

    def train_fasttext(self):
        stu_ids = self.train_dict['stu_id']
        exer_ids = self.train_dict['exer_seq']
        label_ids = self.train_dict['label_seq']
        mask_ids = self.train_dict['mask_seq']
        len_s = len(stu_ids)
        ex_stuid = -1
        all_str = []
        for i in range(len_s):
            tmp_stu_id = stu_ids[i]
            if ex_stuid != tmp_stu_id:
                ex_stuid = tmp_stu_id
                if i !=0:
                    all_str.append(s_str)
                s_str = []
            tmp_exer = exer_ids[i, :]
            tmp_label = label_ids[i, :]
            tmp_mask = mask_ids[i, :]
            for mk in range(len(tmp_mask)):
                if tmp_mask[mk] != 0:  # 有效
                    tmp_str = str(tmp_exer[mk]) + str(int(tmp_label[mk]))+" "
                    s_str.append(tmp_str)
        return all_str

    def forward(self, exer_seq, label_seq, **kwargs):
        input_x = self.exer_emb(exer_seq + label_seq.long() * self.n_item)
        output, _ = self.seq_model(input_x)
        output = self.dropout_layer(output)
        y_pd = self.fc_layer(output).sigmoid()
        return y_pd
    
    @torch.no_grad()
    def predict(self, **kwargs):
        y_pd = self(**kwargs)
        y_pd = y_pd[:, :-1].gather(
            index=kwargs['exer_seq'][:, 1:].unsqueeze(dim=-1), dim=2
        ).squeeze(dim=-1)
        y_pd = y_pd[kwargs['mask_seq'][:, 1:] == 1]
        y_gt = None
        if kwargs.get('label_seq', None) is not None:
            y_gt = kwargs['label_seq'][:, 1:]
            y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        return {
            'y_pd': y_pd,
            'y_gt': y_gt
        }

    def get_main_loss(self, **kwargs):
        y_pd = self(**kwargs)
        y1 = y_pd[:, -1, :].unsqueeze(1)
        y2 = y_pd[:, -1, :].unsqueeze(2)
        ls1 = torch.matmul(y1, self.laplacian)
        ls2 = torch.bmm(ls1, y2).squeeze(-1).squeeze(-1)
        loss_reg = torch.mean(ls2)
        y_pd = y_pd[:, :-1].gather(
            index=kwargs['exer_seq'][:, 1:].unsqueeze(dim=-1), dim=2
        ).squeeze(dim=-1)
        y_pd = y_pd[kwargs['mask_seq'][:, 1:] == 1]
        y_gt = kwargs['label_seq'][:, 1:]
        y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        loss = F.binary_cross_entropy(
            input=y_pd, target=y_gt
        )
        return {
            'loss_reg':loss_reg*self.w_reg,
            'loss_main': loss
        }

    def get_loss_dict(self, **kwargs):
        return self.get_main_loss(**kwargs)



import numpy as np

from .KTInterExtendsQDataTPL import KTInterExtendsQDataTPL
import torch


class DIMKTDataTPL(KTInterExtendsQDataTPL):
    default_cfg = {
        'mid2cache_op_seq': ['M2C_Label2Int', 'M2C_ReMapId', 'M2C_GenUnFoldKCSeq', 'M2C_BuildSeqInterFeats', 'M2C_RandomDataSplit4KT', 'M2C_GenKCSeq', "M2C_DIMKT_OP"],
        'M2C_BuildSeqInterFeats': {
            # 'window_size': 200,
            "extra_inter_feats": ['start_timestamp:float', 'cpt_unfold:token']
        }
    }

    def __getitem__(self, index):
        dic = super().__getitem__(index)
        dic['qd_seq'] = np.stack(
            [self.q_dif[exer_seq][0] for exer_seq in dic['exer_seq']], axis=0
        )
        dic['cd_seq'] = np.stack(
            [self.c_dif[cpt_seq][0] for cpt_seq in dic['cpt_unfold_seq']], axis=0
        )
        dic['cd_seq'] = np.squeeze(dic['cd_seq'])
        mask = dic['mask_seq']==0
        dic['qd_seq'][mask]=0
        dic['cd_seq'][mask] = 0
        return dic

    def set_info_for_fold(self, fold_id):
        super().set_info_for_fold(fold_id)
        self.q_dif = self.final_kwargs['q_diff_list'][fold_id]
        self.c_dif = self.final_kwargs['c_diff_list'][fold_id]



# -*- coding: utf-8 -*-
# @Author : Xiangzhi Chen
# @Github : kervias

import logging
import pytz
import sys
import os


class Logger(object):
    def __init__(self, filepath: str = None,
                 fmt: str = "%(asctime)s[%(levelname)s]: %(message)s",
                 date_fmt: str = "%Y-%m-%d %H:%M:%S",
                 timezone: str = "Asia/Shanghai",
                 level=logging.DEBUG,
                 DISABLE_LOG_STDOUT=False
                 ):
        self.timezone = timezone
        if filepath:
            dir_name = os.path.dirname(filepath)
            if not os.path.exists(dir_name):
                os.makedirs(dir_name)

        self.logger = logging.getLogger("edustudio")
        self.logger.setLevel(level)
        formatter = logging.Formatter(
            fmt, datefmt=date_fmt
        )
        formatter.converter = self.converter

        if filepath:
            # write into file
            fh = logging.FileHandler(filepath, mode='a+', encoding='utf-8')
            fh.setLevel(level)
            fh.setFormatter(formatter)
            self.logger.addHandler(fh)

        # show on console
        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(level)
        ch.setFormatter(formatter)
        flag = False
        for handler in self.logger.handlers:
            if type(handler) is logging.StreamHandler:
                flag = True
        if flag is False and DISABLE_LOG_STDOUT is False:
            self.logger.addHandler(ch)

    def _flush(self):
        for handler in self.logger.handlers:
            handler.flush()

    def debug(self, message):
        self.logger.debug(message)
        self._flush()

    def info(self, message):
        self.logger.info(message)
        self._flush()

    def warning(self, message):
        self.logger.warning(message)
        self._flush()

    def error(self, message):
        self.logger.error(message)
        self._flush()

    def critical(self, message):
        self.logger.critical(message)
        self._flush()

    def converter(self, sec):
        tz = pytz.timezone(self.timezone)
        dt = pytz.datetime.datetime.fromtimestamp(sec, tz)
        return dt.timetuple()

    def get_std_logger(self):
        return self.logger



r"""
MGCD
##################################
Reference:
    Huang et al. "Group-level cognitive diagnosis: A multi-task learning perspective." in ICDM 2021.
"""
import numpy as np
from ..gd_basemodel import GDBaseModel
import torch.nn as nn
from ..utils.components import PosMLP
import torch
import torch.nn.functional as F


class MGCD(GDBaseModel):
    """
    prednet_len1: The dimension of the first hidden layer of the fully connected layer before predicting the score
    prednet_len2: The dimension of the second hidden layer of the fully connected layer before predicting the score
    """
    default_cfg = {
        'prednet_len1': 128,
        'prednet_len2': 64,
    }
    def __init__(self, cfg):
        super().__init__(cfg)

    def add_extra_data(self, inter_student, df_G, **kwargs):
        self.stu_n = inter_student['stu_id'].max()
        self.df_G = df_G[['stu_id', 'group_id']].set_index('stu_id')['group_id'].to_dict()  # stu2group

        group2stu = {}
        self.stu_n = max(self.df_G, default=self.stu_n)
        for key, value in self.df_G.items():
            group2stu.setdefault(value, []).append(key)

        self.stu_n += 1
        self.df_G = group2stu
        self.stu_exe2label = inter_student.set_index(['stu_id', 'exer_id'])['label'].to_dict()
        self.Q_mat = kwargs['Q_mat'].to(self.device)

        self.stu2exe = {}
        for stu_id, exer_id in zip(map(int, inter_student['stu_id']), map(int, inter_student['exer_id'])):
            self.stu2exe.setdefault(stu_id, []).append(exer_id)

    def build_cfg(self):
        self.group_n = self.datatpl_cfg['dt_info']['group_count']
        self.exer_n = self.datatpl_cfg['dt_info']['exer_count']
        self.knowledge_dim = self.datatpl_cfg['dt_info']['cpt_count']
        self.stu_dim = self.knowledge_dim
        self.prednet_input_len = self.knowledge_dim
        self.prednet_len1 = self.modeltpl_cfg['prednet_len1']
        self.prednet_len2 = self.modeltpl_cfg['prednet_len2']
        self.num_hidden = self.stu_dim

    def build_model(self):
        self.R = nn.Embedding(self.stu_n, self.stu_dim)
        # h
        self.h = nn.Parameter(torch.randn(self.stu_dim, 1, requires_grad=True))
        # W_k
        self.W_k = nn.Parameter(torch.randn(self.num_hidden, self.stu_dim, requires_grad=True))
        # W_q
        self.W_q = nn.Parameter(torch.randn(self.num_hidden, self.stu_dim, requires_grad=True))
        # W_c
        self.W_c = nn.Embedding(self.group_n, self.stu_dim)

        # A
        self.A = nn.Parameter(torch.randn(self.stu_dim, self.stu_dim, requires_grad=True))
        # B
        self.B = nn.Embedding(self.exer_n, self.knowledge_dim)
        # D
        self.D = nn.Embedding(self.exer_n, 1)

        self.prednet_full1 = nn.Linear(self.prednet_input_len, self.prednet_len1)
        self.drop_1 = nn.Dropout(p=0.5)
        self.prednet_full2 = nn.Linear(self.prednet_len1, self.prednet_len2)
        self.drop_2 = nn.Dropout(p=0.5)
        self.prednet_full3 = nn.Linear(self.prednet_len2, 1)

    def _init_params(self):
        for name, param in self.named_parameters():
            if 'weight' in name:
                nn.init.xavier_normal_(param)

    def forward(self, group_id, exer_id, **kwargs):  # 这里应该得到group对exercise作答正确的概率
        # before prednet
        self.apply_clipper()

        """
        :param group_id: LongTensor
        :param group_member: list of LongTensor
        :param exer_ids: LongTensor
        :param kn_emb: FloatTensor, the knowledge relevancy vectors  batch_size*knowledge_dim
        :return: FloatTensor, the probabilities of answering correctly
        """
        group_member = [self.df_G[int(group_id[i])] for i in range(len(group_id))]
        kn_emb = self.Q_mat[exer_id]

        group_emb_list = []
        for i in range(len(group_member)):
            stu_emb = self.R(torch.tensor(group_member[i]).to(self.device))
            ci = self.W_c(group_id[i])
            a = torch.matmul(stu_emb, self.W_k)
            b = torch.mv(self.W_q, ci)
            oj = torch.tanh(a + b)
            oj = torch.matmul(self.h.T, oj.T)
            oj = torch.softmax(oj, dim=1)
            group_emb = torch.matmul(oj, stu_emb)
            group_emb_list.append(group_emb)

        group_emb = torch.cat(group_emb_list)

        group_emb = torch.sigmoid(torch.matmul(group_emb, self.A))
        k_difficulty = torch.sigmoid(self.B(exer_id))
        e_discrimination = torch.sigmoid(self.D(exer_id)) * 10

        input_x = e_discrimination * (group_emb - k_difficulty) * kn_emb
        input_x = self.drop_1(torch.tanh(self.prednet_full1(input_x)))
        input_x = self.drop_2(torch.tanh(self.prednet_full2(input_x)))
        output = torch.sigmoid(self.prednet_full3(input_x))

        return output

    def apply_clipper(self):
        clipper = NoneNegClipper()
        self.prednet_full1.apply(clipper)
        self.prednet_full2.apply(clipper)
        self.prednet_full3.apply(clipper)

    @torch.no_grad()
    def predict(self, group_id, exer_id, **kwargs):
        return {
            'y_pd': self(group_id, exer_id).flatten(),
        }

    def get_main_loss(self, **kwargs):
        # 这里的loss跟论文中loss对应
        group_id = kwargs['group_id']
        exer_id = kwargs['exer_id']
        label = kwargs['label']
        pd = self(group_id, exer_id).flatten()
        loss_mse = nn.MSELoss(reduction = 'mean')
        loss_group = loss_mse(pd, label)

        stu_ids = []
        for i in range(len(group_id)):
            stu_ids = stu_ids + self.df_G[int(group_id[i])]
        stu_id = []
        exer_stu = []
        label_stu = []
        for stu in stu_ids:
            if stu in self.stu2exe:
                for exer in self.stu2exe[stu]:
                    stu_id.append(stu)
                    exer_stu.append(exer)
                    label_stu.append(self.stu_exe2label[(stu, exer)])
        if stu_id != []:
            stu_id = torch.tensor(stu_id).to(self.device)
            exer_id = torch.tensor(exer_stu).to(self.device)
            label_stu = torch.unsqueeze(torch.tensor(label_stu), dim=1).to(self.device)
            kn_emb = self.Q_mat[exer_id]
            stu_emb = self.R(stu_id)
            stu_emb = torch.sigmoid(torch.matmul(stu_emb, self.A))
            k_difficulty = torch.sigmoid(self.B(exer_id))
            e_discrimination = torch.sigmoid(self.D(exer_id)) * 10
            # prednet
            input_x = e_discrimination * (stu_emb - k_difficulty) * kn_emb
            input_x = self.drop_1(torch.tanh(self.prednet_full1(input_x)))
            input_x = self.drop_2(torch.tanh(self.prednet_full2(input_x)))
            output = torch.sigmoid(self.prednet_full3(input_x)).to(self.device)

            loss_stu = F.binary_cross_entropy(input=output, target=label_stu)
        else:
            loss_stu = 0

        return {
            'loss_main': loss_group + loss_stu
        }

    def get_loss_dict(self, **kwargs):
        return self.get_main_loss(**kwargs)
    
    def get_stu_status(self, stu_id=None):
        if stu_id is not None:
            return self.W_c(stu_id)
        else:
            return self.W_c.weight


class NoneNegClipper(object):
    """
    Make the parameters of the fully connected layer non-negative
    """
    def __init__(self):
        super(NoneNegClipper, self).__init__()

    def __call__(self, module):
        if hasattr(module, 'weight'):
            w = module.weight.data
            a = torch.relu(torch.neg(w))
            w.add_(a)




r"""
CNCD-F
##########################################

Reference:
    Fei Wang et al. "Neural Cognitive Diagnosis for Intelligent Education Systems" in AAAI 2020.

Reference Code:
    https://github.com/bigdata-ustc/Neural_Cognitive_Diagnosis-NeuralCD

"""
from ..gd_basemodel import GDBaseModel
import torch.nn as nn
from ..utils.components import PosMLP
import torch
import torch.nn.functional as F
import numpy as np


class TextCNN(nn.Module):
    def __init__(self, batch_size, knowledge_n, embedding_dim):
        self.batch_size = batch_size
        self.embedding_len = embedding_dim
        self.sequence_len = 600
        self.output_len = knowledge_n
        self.channel_num1, self.channel_num2, self.channel_num3 = 400, 200, 100
        self.kernel_size1, self.kernel_size2, self.kernel_size3 = 3, 4, 5
        self.pool1 = 3
        self.full_in = (self.sequence_len + self.kernel_size1 - 1) // self.pool1 + self.kernel_size2 + self.kernel_size3 - 2
        super(TextCNN, self).__init__()

        self.conv1 = nn.Conv1d(self.embedding_len, self.channel_num1, kernel_size=self.kernel_size1, padding=self.kernel_size1-1, stride=1)
        self.conv2 = nn.Conv1d(self.channel_num1, self.channel_num2, kernel_size=self.kernel_size2, padding=self.kernel_size2-1, stride=1)
        self.conv3 = nn.Conv1d(self.channel_num2, self.channel_num3, kernel_size=self.kernel_size3, padding=self.kernel_size3 - 1, stride=1)
        self.full = nn.Linear(self.full_in, embedding_dim)

    def prepare_embedding(self, content_list):
        from gensim.models import Word2Vec
        from bintrees import FastRBTree
        words_all = []
        for v in content_list:
            ls = [str(i) for i in v.split(',')]
            words_all.extend(ls)
        word2vec_model = Word2Vec(words_all)
        words, self.word2id = FastRBTree(), FastRBTree()
        word_npy = []
        for exer in words_all:
            for word in exer:
                words.insert(word, True)
        word_count = 0
        word_npy.append([0.] * 100)  # index=0 is a zero-vector
        for word in words:
            if word in word2vec_model.wv:
                word_count += 1
                self.word2id.insert(word, word_count)
                word_npy.append(word2vec_model.wv[word])
            else:
                print('not found: ' + str(word))
        self.word_npy = np.array(word_npy)
        self.word_emb = nn.Embedding(len(self.word2id) + 1, self.embedding_len, padding_idx=0)
        self.word_emb.weight.data.copy_(torch.from_numpy(self.word_npy))
        return self.word2id

    def forward(self, x):
        x = self.word_emb(x)
        x = torch.transpose(x, 1, 2)
        x = F.relu(self.conv1(x))
        x = F.max_pool1d(x, kernel_size=self.pool1)
        x = F.relu(self.conv2(x))
        x = F.relu(self.conv3(x))
        x = torch.transpose(x, 1, 2)
        x = F.max_pool1d(x, self.channel_num3)
        x = torch.transpose(x, 1, 2).view(-1, self.full_in)
        ret = self.full(x)     # 使用的损失函数包含sigmoid，在预测时需在网络外加sigmoid
        return ret

class CNCD_F(GDBaseModel):
    r"""
    CNCD-F

    default_cfg:
       'dnn_units': [512, 256]  # dimension list of hidden layer in prediction layer
       'dropout_rate': 0.5      # dropout rate
       'activation': 'sigmoid'  # activation function in prediction layer
       'disc_scale': 10         # discrimination scale
       'max_len': 600,          # the maximum length of the exercise text
        'text_embedding_dim': 100 # dimension of text embedding
    """
    default_cfg = {
        'dnn_units': [512, 256],
        'dropout_rate': 0.5,
        'activation': 'sigmoid',
        'disc_scale': 10,
        'max_len': 600,
        'text_embedding_dim': 100
    }
    def __init__(self, cfg):
        super().__init__(cfg)

    def build_cfg(self):
        self.n_user = self.datatpl_cfg['dt_info']['stu_count']
        self.n_item = self.datatpl_cfg['dt_info']['exer_count']
        self.n_cpt = self.datatpl_cfg['dt_info']['cpt_count']
        self.text_dim = self.modeltpl_cfg["text_embedding_dim"]
        self.max_len = self.modeltpl_cfg["max_len"]

    def get_from_text(self):
        x = []
        for e in self.content_list:
            word_ids = []
            for word in e[:self.max_len]:  # 固定长度为self.max_len
                word_id = self.word2id.get(word)
                if word_id is not None:
                    word_ids.append(word_id)
            if len(word_ids) < self.max_len:
                word_ids += [0] * (self.max_len - len(word_ids))  # padding到self.max_len
            x.append(word_ids)
        return torch.tensor(x, device=self.traintpl_cfg['device'])

    def build_model(self):
        # prediction sub-net
        self.textcnn = TextCNN(self.traintpl_cfg["batch_size"], self.n_cpt, self.text_dim)
        self.word2id = self.textcnn.prepare_embedding(self.content_list)
        self.word_ids = self.get_from_text()
        self.out_text_factor = nn.Linear(self.text_dim, 1)
        self.student_emb = nn.Embedding(self.n_user, self.n_cpt+1)
        self.k_difficulty = nn.Embedding(self.n_item, self.n_cpt)
        self.e_difficulty = nn.Embedding(self.n_item, 1)
        self.pd_net = PosMLP(
            input_dim=self.n_cpt+1, output_dim=1, activation=self.modeltpl_cfg['activation'],
            dnn_units=self.modeltpl_cfg['dnn_units'], dropout_rate=self.modeltpl_cfg['dropout_rate']
        )

    def add_extra_data(self, **kwargs):
        cl = kwargs['content']
        self.content_list=[]
        for ls in cl:
            s = ",".join(map(str, ls))
            self.content_list.append(s)
        self.Q_mat = kwargs['Q_mat']

    def forward(self, stu_id, exer_id):
        # before prednet
        items_Q_mat = self.Q_mat[exer_id].to(self.traintpl_cfg['device'])
        items_content = self.word_ids[exer_id]
        text_embedding = self.textcnn(items_content)
        text_factor = torch.sigmoid(self.out_text_factor(text_embedding))
        stu_emb = self.student_emb(stu_id)
        stat_emb = torch.sigmoid(stu_emb)
        k_difficulty = torch.sigmoid(self.k_difficulty(exer_id))
        k_difficulty = torch.cat((k_difficulty, text_factor), dim=1)
        e_difficulty = torch.sigmoid(self.e_difficulty(exer_id)) * self.modeltpl_cfg['disc_scale']
        # prednet
        text_factor_q = torch.ones((items_Q_mat.shape[0], 1), device=self.traintpl_cfg['device'])
        input_knowledge_point = torch.cat((items_Q_mat, text_factor_q), dim=1)
        input_x = e_difficulty * (stat_emb - k_difficulty) * input_knowledge_point
        pd = self.pd_net(input_x).sigmoid()
        return pd

    @torch.no_grad()
    def predict(self, stu_id, exer_id, **kwargs):
        return {
            'y_pd': self(stu_id, exer_id).flatten(),
        }

    def get_main_loss(self, **kwargs):
        stu_id = kwargs['stu_id']
        exer_id = kwargs['exer_id']
        label = kwargs['label']
        pd = self(stu_id, exer_id).flatten()
        loss = F.binary_cross_entropy(input=pd, target=label)
        return {
            'loss_main': loss
        }

    def get_loss_dict(self, **kwargs):
        return self.get_main_loss(**kwargs)
    
    def get_stu_status(self, stu_id=None):
        if stu_id is not None:
            return self.student_emb(stu_id)
        else:
            return self.student_emb.weight



r"""
EERNN
##########################################

Reference:
    Yu Su et al. "Exercise-Enhanced Sequential Modeling for Student Performance Prediction" in AAAI 2018.

Reference Code:
    https://github.com/shaoliangliang1996/EERNN

"""

from ..gd_basemodel import GDBaseModel
import torch.nn as nn
import torch
import torch.nn.functional as F


class EERNNM(GDBaseModel):
    default_cfg = {
        'd_v': 100,
        'd_h': 100,
    }
    
    def add_extra_data(self, **kwargs):
        self.w2v_text_emb = kwargs['w2v_word_emb']
        self.exer_content = kwargs['exer_content'].to(self.device)

    def build_cfg(self):
        self.n_user = self.datatpl_cfg['dt_info']['stu_count']
        self.n_item = self.datatpl_cfg['dt_info']['exer_count']
        self.n_word = self.datatpl_cfg['dt_info']['word_count']
        self.d_0 = self.datatpl_cfg['dt_info']['word_emb_dim']
        self.d_v = self.modeltpl_cfg['d_v']
        self.d_h = self.modeltpl_cfg['d_h']

    def build_model(self):
        self.word_emb = nn.Embedding(self.n_word, self.d_0, padding_idx=0)
        self.exer_text_seq_model = nn.LSTM(self.d_0, self.d_v, batch_first=True, bidirectional=True)
        self.stu_inter_seq_model = nn.LSTM(4 * self.d_v, self.d_h, batch_first=True)
        self.pd_layer = nn.Sequential(
            nn.Linear(in_features=self.d_h + 2*self.d_v, out_features=(self.d_h + 2*self.d_v) // 2),
            nn.ReLU(inplace=True),
            nn.Linear(in_features=(self.d_h + 2*self.d_v) // 2, out_features=1),
            nn.Sigmoid()
        )

    def _init_params(self):
        super()._init_params()
        self.word_emb.weight.data.copy_(torch.from_numpy(self.w2v_text_emb))

    def get_exer_embedding(self, exers=None):
        if exers is not None:
            t = self.word_emb(self.exer_content[exers])
        else:
            t = self.word_emb(self.exer_content)
        mix_h, _ = self.exer_text_seq_model(t)
        forward_h, reverse_h = torch.chunk(mix_h, 2, dim=-1)
        v_m = torch.cat([forward_h, reverse_h.flip(dims=[0])], dim=-1)
        return torch.max(v_m, dim=1)[0]

    def forward(self, exer_seq, label_seq, **kwargs):
        exers, idx = exer_seq.unique(return_inverse=True)
        exer_emb = self.get_exer_embedding(exers)[idx]
        fused_exer_emb = exer_emb.repeat(1,1,2)
        cond = torch.where(label_seq == 1)
        fused_exer_emb[cond[0], cond[1], self.d_v*2:] = 0.0
        cond = torch.where(label_seq == 0)
        fused_exer_emb[cond[0], cond[1], 0:self.d_v*2] = 0.0

        h, _ = self.stu_inter_seq_model(fused_exer_emb)
        feat = torch.cat([h[:,0:-1,:], exer_emb[:,1:,:]], dim=-1)
        return self.pd_layer(feat)

    @torch.no_grad()
    def predict(self, **kwargs):
        y_pd = self(**kwargs).squeeze(dim=-1)
        y_pd = y_pd[kwargs['mask_seq'][:, 1:] == 1]
        y_gt = None
        if kwargs.get('label_seq', None) is not None:
            y_gt = kwargs['label_seq'][:, 1:]
            y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        return {
            'y_pd': y_pd,
            'y_gt': y_gt
        }

    def get_main_loss(self, **kwargs):
        y_pd = self(**kwargs).squeeze(dim=-1)
        y_pd = y_pd[kwargs['mask_seq'][:, 1:] == 1]
        y_gt = kwargs['label_seq'][:, 1:]
        y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        loss = F.binary_cross_entropy(
            input=y_pd, target=y_gt
        )
        return {
            'loss_main': loss
        }

    def get_loss_dict(self, **kwargs):
        return self.get_main_loss(**kwargs)


class EERNNA(EERNNM):
    def forward(self, exer_seq, label_seq, **kwargs):
        exers, idx = exer_seq.unique(return_inverse=True)
        exer_emb_unique = self.get_exer_embedding(exers)
        exer_emb = exer_emb_unique[idx]
        fused_exer_emb = exer_emb.repeat(1,1,2)
        cond = torch.where(label_seq == 1)
        fused_exer_emb[cond[0], cond[1], self.d_v*2:] = 0.0
        cond = torch.where(label_seq == 0)
        fused_exer_emb[cond[0], cond[1], 0:self.d_v*2] = 0.0

        h, _ = self.stu_inter_seq_model(fused_exer_emb)
        
        # 加权h

        # a_norm = exer_emb_unique / exer_emb_unique.norm(dim=1)[:, None]
        # cos_sim_mat = torch.mm(a_norm, a_norm.transpose(0,1)) # cos similarity
        h_new = []
        for tid in range(exer_seq.shape[-1] - 1):
            cos_sim = F.cosine_similarity(exer_emb[:, [tid+1], :], exer_emb[:, 0:tid+1, :], dim=-1)
            h_new.append((h[:,0:tid+1, :] *  cos_sim.unsqueeze(dim=-1)).sum(dim=1, keepdim=True))

        h_new = torch.cat(h_new, dim=1)
        feat = torch.cat([h_new, exer_emb[:,1:,:]], dim=-1)
        return self.pd_layer(feat)



from edustudio.traintpl import GeneralTrainTPL
from edustudio.utils.common import tensor2npy
import torch
from tqdm import tqdm


class GroupCDTrainTPL(GeneralTrainTPL):
    
    @torch.no_grad()
    def evaluate(self, loader):
        self.model.eval()
        stu_id_list = list(range(len(loader)))
        pd_list = list(range(len(loader)))
        gt_list = list(range(len(loader)))
        for idx, batch_dict in enumerate(tqdm(loader, ncols=self.frame_cfg['TQDM_NCOLS'], desc="[PREDICT]")):
            batch_dict = self.batch_dict2device(batch_dict)
            eval_dict = self.model.predict(**batch_dict)
            stu_id_list[idx] = batch_dict['group_id']
            pd_list[idx] = eval_dict['y_pd']
            gt_list[idx] = eval_dict['y_gt'] if 'y_gt' in eval_dict else batch_dict['label']
        y_pd = torch.hstack(pd_list)
        y_gt = torch.hstack(gt_list)
        group_id = torch.hstack(stu_id_list)

        eval_data_dict = {
            'group_id': group_id,
            'y_pd': y_pd,
            'y_gt': y_gt,
        }
        if hasattr(self.model, 'get_stu_status'):
            stu_stats_list = []
            idx = torch.arange(0, self.datatpl_cfg['dt_info']['group_count']).to(self.traintpl_cfg['device'])
            for i in range(0,self.datatpl_cfg['dt_info']['group_count'], self.traintpl_cfg['eval_batch_size']):
                batch_stu_id = idx[i:i+self.traintpl_cfg['eval_batch_size']]
                batch = self.model.get_stu_status(batch_stu_id)
                stu_stats_list.append(batch)
            stu_stats = torch.vstack(stu_stats_list)
            eval_data_dict.update({
                'stu_stats': tensor2npy(stu_stats),
            })
        if hasattr(self.datatpl, 'Q_mat'):
            eval_data_dict.update({
                'Q_mat': tensor2npy(self.datatpl.Q_mat)
            })
        eval_result = {}
        for evaltpl in self.evaltpls: eval_result.update(
                evaltpl.eval(ignore_metrics=self.traintpl_cfg['ignore_metrics_in_train'], **eval_data_dict)
            )
        return eval_result

    @torch.no_grad()
    def inference(self, loader):
        self.model.eval()
        stu_id_list = list(range(len(loader)))
        pd_list = list(range(len(loader)))
        gt_list = list(range(len(loader)))
        for idx, batch_dict in enumerate(tqdm(loader, ncols=self.frame_cfg['TQDM_NCOLS'], desc="[PREDICT]")):
            batch_dict = self.batch_dict2device(batch_dict)
            eval_dict = self.model.predict(**batch_dict)
            stu_id_list[idx] = batch_dict['group_id']
            pd_list[idx] = eval_dict['y_pd']
            gt_list[idx] = eval_dict['y_gt'] if 'y_gt' in eval_dict else batch_dict['label']
        y_pd = torch.hstack(pd_list)
        y_gt = torch.hstack(gt_list)
        group_id = torch.hstack(stu_id_list)

        eval_data_dict = {
            'group_id': group_id,
            'y_pd': y_pd,
            'y_gt': y_gt,
        }
        if hasattr(self.model, 'get_stu_status'):
            stu_stats_list = []
            idx = torch.arange(0, self.datatpl_cfg['dt_info']['group_count']).to(self.traintpl_cfg['device'])
            for i in range(0,self.datatpl_cfg['dt_info']['group_count'], self.traintpl_cfg['eval_batch_size']):
                batch_stu_id = idx[i:i+self.traintpl_cfg['eval_batch_size']]
                batch = self.model.get_stu_status(batch_stu_id)
                stu_stats_list.append(batch)
            stu_stats = torch.vstack(stu_stats_list)
            eval_data_dict.update({
                'stu_stats': tensor2npy(stu_stats),
            })
        if hasattr(self.datatpl, 'Q_mat'):
            eval_data_dict.update({
                'Q_mat': tensor2npy(self.datatpl.Q_mat)
            })
        eval_result = {}
        for evaltpl in self.evaltpls: eval_result.update(evaltpl.eval(**eval_data_dict))
        return eval_result



from .base_mid2cache import BaseMid2Cache
import numpy as np
import pandas as pd
from itertools import chain
import torch
from edustudio.utils.common import set_same_seeds


class M2C_BuildMissingQ(BaseMid2Cache):
    default_cfg = {
        'seed': 20230518,
        'Q_delete_ratio': 0.0,
    }

    def process(self, **kwargs):
        dt_info = kwargs['dt_info']
        self.item_count = dt_info['exer_count']
        self.cpt_count = dt_info['cpt_count']
        self.df_Q = kwargs['df_exer'][['exer_id:token', 'cpt_seq:token_seq']]

        self.missing_df_Q = self.get_missing_df_Q()
        self.missing_Q_mat = self.get_Q_mat_from_df_arr(self.missing_df_Q, self.item_count, self.cpt_count)

        kwargs['missing_df_Q'] = self.missing_df_Q
        kwargs['missing_Q_mat'] = self.missing_Q_mat

        return kwargs

    def get_missing_df_Q(self):
        set_same_seeds(seed=self.m2c_cfg['seed'])
        ratio = self.m2c_cfg['Q_delete_ratio']
        iid2cptlist = self.df_Q.set_index('exer_id:token')['cpt_seq:token_seq'].to_dict()
        iid_lis = np.array(list(chain(*[[i]*len(iid2cptlist[i]) for i in iid2cptlist])))
        cpt_lis = np.array(list(chain(*list(iid2cptlist.values()))))
        entry_arr = np.vstack([iid_lis, cpt_lis]).T

        np.random.shuffle(entry_arr)

        # reference: https://stackoverflow.com/questions/64834655/python-how-to-find-first-duplicated-items-in-an-numpy-array
        _, idx = np.unique(entry_arr[:, 1], return_index=True) # 先从每个知识点中选出1题出来
        bool_idx = np.zeros_like(entry_arr[:, 1], dtype=bool)
        bool_idx[idx] = True
        preserved_exers = np.unique(entry_arr[bool_idx, 0]) # 选择符合条件的习题作为保留

        delete_num = int(ratio * self.item_count)
        preserved_num = self.item_count - delete_num

        if len(preserved_exers) >= preserved_num:
            self.logger.warning(
                f"Cant Satisfy Delete Require: {len(preserved_exers)=},{preserved_num=}"
            )
        else:
            need_preserved_num = preserved_num - len(preserved_exers)

            left_iids = np.arange(self.item_count)
            left_iids = left_iids[~np.isin(left_iids, preserved_exers)]
            np.random.shuffle(left_iids)
            choose_iids = left_iids[0:need_preserved_num]

            preserved_exers = np.hstack([preserved_exers, choose_iids])

        return self.df_Q.copy()[self.df_Q['exer_id:token'].isin(preserved_exers)].reset_index(drop=True)


    def get_Q_mat_from_df_arr(self, df_Q_arr, item_count, cpt_count):
        Q_mat = torch.zeros((item_count, cpt_count), dtype=torch.int64)
        for _, item in df_Q_arr.iterrows(): Q_mat[item['exer_id:token'], item['cpt_seq:token_seq']] = 1
        return Q_mat



from ..common import BaseMid2Cache
import numpy as np
from collections import defaultdict


class M2C_DKTForget_OP(BaseMid2Cache):
    default_cfg = {
        'timestamp_unit': 'ms'
    }

    def process(self, **kwargs):
        df_train_folds = kwargs['df_train_folds']
        df_valid_folds = kwargs['df_valid_folds']
        df_test_folds = kwargs['df_test_folds']

        self.dt_info = kwargs['dt_info']
        self.dt_info['n_pcount_list'] = []
        self.dt_info['n_rgap_list'] = []
        self.dt_info['n_sgap_list'] = []

        for idx, (train_dict, test_dict) in enumerate(zip(df_train_folds, df_test_folds)):
            self.train_dict = train_dict
            self.test_dict = test_dict
            if df_valid_folds is not None and len(df_valid_folds) > 0:
                self.val_dict = df_valid_folds[idx]
                
            self.n_pcount, self.n_rgap, self.n_sgap = 0, 0, 0
            self._construct_s_gap(self.train_dict)
            self._construct_r_gap_and_p_count(self.train_dict)
            if self.val_dict:
                self._construct_s_gap(self.val_dict)
                self._construct_r_gap_and_p_count(self.val_dict)
            self._construct_s_gap(self.test_dict)
            self._construct_r_gap_and_p_count(self.test_dict)

            self.dt_info['n_pcount_list'].append(self.n_pcount)
            self.dt_info['n_rgap_list'].append(self.n_rgap)
            self.dt_info['n_sgap_list'].append(self.n_sgap)
        
        return kwargs

    @staticmethod
    def time2minutes(tensor, curr_unit):
        if curr_unit == 'ms':
            return tensor // (1000 * 60)
        elif curr_unit == 's':
            return tensor // 60
        elif curr_unit == 'minute':
            return tensor
        else:
            ValueError(f"unsupported answer_time_unit: {curr_unit}")

    def _construct_s_gap(self, data_dict):
        time_seq = data_dict['start_timestamp_seq:float_seq']
        mask_seq = data_dict['mask_seq:token_seq']
        s_gap = self.time2minutes(time_seq - np.hstack([time_seq[:,[0]], time_seq[:,0:-1]]), curr_unit=self.m2c_cfg['timestamp_unit'])
        s_gap[mask_seq == 0] = 1.0 # 防止log报错
        s_gap[:, 0] = 1.0 # 防止log报错
        assert (s_gap < 0).sum() == 0 # 确保时间是顺序的
        s_gap[s_gap == 0] = 1.0 # 防止log报错
        s_gap = np.log2(s_gap).astype(np.int64) + 1
        s_gap[mask_seq == 0] = 0 
        s_gap[:, 0] = 0 
        data_dict['s_gap'] = s_gap
        self.n_sgap = max(self.n_sgap, s_gap.max() + 1)

    def _construct_r_gap_and_p_count(self, data_dict):
        time_seq = data_dict['start_timestamp_seq:float_seq']
        mask_seq = data_dict['mask_seq:token_seq']
        exer_seq = data_dict['exer_seq:token_seq']

        p_count_mat = np.zeros(mask_seq.shape, dtype=np.float32)
        r_gap_mat = np.zeros(mask_seq.shape, dtype=np.float32)
        for seq_id, exer_seq_one in enumerate(exer_seq):
            exer_seq_mask_one = mask_seq[seq_id]
            time_seq_one = time_seq[seq_id]
            p_count_dict = defaultdict(int) # 记录习题被回答次数
            r_gap_dict = {} # 记录习题上次被回答时的时间
            for w_id, exer_id in enumerate(exer_seq_one[exer_seq_mask_one == 1]):
                exer_id = exer_id
                p_count_mat[seq_id, w_id] = p_count_dict[exer_id]
                p_count_dict[exer_id] += 1
                if exer_id in r_gap_dict:
                    r_gap_mat[seq_id, w_id] = self.time2minutes(time_seq_one[w_id] - r_gap_dict[exer_id],  curr_unit=self.m2c_cfg['timestamp_unit']) # 变为分钟
                    r_gap_dict[exer_id] = time_seq_one[w_id]
                else:
                    r_gap_dict[exer_id] = time_seq_one[w_id]
                    r_gap_mat[seq_id, w_id] = 0

        p_count_mat[:, 0] = 1.0
        p_count_mat[mask_seq == 0] = 1.0
        p_count_mat[p_count_mat == 0] = 1.0
        p_count_mat = np.log2(p_count_mat).astype(np.int64) + 1
        p_count_mat[:,0] = 0 # 这些单独作为一类
        p_count_mat[mask_seq == 0] = 0  # 这些单独作为一类
        
        r_gap_mat[:,0] = 1.0
        r_gap_mat[mask_seq == 0] = 1.0
        r_gap_mat[r_gap_mat == 0] = 1.0
        r_gap_mat = np.log2(r_gap_mat).astype(np.int64) + 1
        r_gap_mat[:,0] = 0
        r_gap_mat[mask_seq == 0] =  0

        data_dict['r_gap'] = r_gap_mat
        data_dict['p_count'] = p_count_mat

        self.n_pcount = max(self.n_pcount, p_count_mat.max()+ 1)
        self.n_rgap = max(self.n_rgap, r_gap_mat.max() + 1)



r"""
CDGK
##################################
Reference:
    Haiping Ma et al. "Using Knowledge Concept Aggregation towards Accurate Cognitive Diagnosis." in CIKM 2021.
"""

from ..gd_basemodel import GDBaseModel
import torch.nn as nn
import torch
from ..utils.components import PosLinear
import torch.nn.functional as F
from collections import defaultdict
import pandas as pd
from edustudio.utils.common import tensor2npy
import numpy as np
from typing import Dict


class CDGK_META(nn.Module):
    def __init__(self, stu_count, exer_count, cpt_count):
        super().__init__()
        self.n_user = stu_count
        self.n_item = exer_count
        self.n_cpt = cpt_count
        self.build_model()

    def build_model(self):
        self.stu_emb = nn.Embedding(self.n_user, self.n_cpt)
        self.exer_emb = nn.Embedding(self.n_item, self.n_cpt)
        self.exer_disc = nn.Embedding(self.n_item, 1)
        self.guess_stu_emb = nn.Embedding(self.n_user, 1)
        self.guess_cpt_emb = nn.Embedding(self.n_cpt, 1)

        self.pred_layer = nn.Sequential(
            PosLinear(self.n_cpt, 1),
            nn.Sigmoid()
        )

        self.guess_pred_layer = nn.Sequential(
            nn.Linear(1, 1),
            nn.Sigmoid()
        )

    def forward(self, stu_id, exer_id, **kwargs):
        items_Q_mat = self.Q_mat[exer_id]
        stu_emb = self.stu_emb(stu_id).sigmoid()
        k_difficulty = self.exer_emb(exer_id).sigmoid()
        e_difficulty = self.exer_disc(exer_id).sigmoid()

        input_x = e_difficulty * (stu_emb - k_difficulty) * items_Q_mat
        pd1 = self.pred_layer(input_x)

        pd2 = self.guess_pred_layer(
            torch.cat(
                [self.guess_stu_emb(stu_id), torch.mm(items_Q_mat.float(), self.guess_cpt_emb.weight)], dim=1
            ).sum(dim=1, keepdim=True)
        )
        return pd2  + (1 - pd2) * pd1

    @torch.no_grad()
    def predict(self, stu_id, exer_id, **kwargs):
        return {
            'y_pd': self(stu_id, exer_id).flatten(),
        }

    def get_main_loss(self, **kwargs):
        stu_id = kwargs['stu_id']
        exer_id = kwargs['exer_id']
        label = kwargs['label']
        pd = self(stu_id, exer_id).flatten()
        loss = F.binary_cross_entropy(input=pd, target=label)
        return {
            'loss_main': loss
        }

    def get_loss_dict(self, **kwargs):
        return self.get_main_loss(**kwargs)
    
    def get_stu_status(self, stu_id=None):
        if stu_id is not None:
            return self.stu_emb(stu_id)
        else:
            return self.stu_emb.weight


class CDGK_SINGLE(GDBaseModel):
    def build_cfg(self):
        self.n_user = self.datatpl_cfg['dt_info']['stu_count']
        self.n_item = self.datatpl_cfg['dt_info']['exer_count']
        self.n_cpt = self.datatpl_cfg['dt_info']['cpt_count']

    def build_model(self):
        self.cdgk = CDGK_META(
            stu_count=self.n_user,
            exer_count=self.n_item,
            cpt_count=self.n_cpt
        )
        self.cdgk.Q_mat = self.Q_mat

    def add_extra_data(self, **kwargs):
        super().add_extra_data(**kwargs)
        self.Q_mat = kwargs['Q_mat'].to(self.device)

    @torch.no_grad()
    def predict(self, **kwargs):
        return self.cdgk.predict(**kwargs)

    def get_main_loss(self, **kwargs):
        return self.cdgk.get_main_loss(**kwargs)

    def get_loss_dict(self, **kwargs):
        return self.get_main_loss(**kwargs)
    
    def get_stu_status(self, stu_id=None):
        return self.cdgk.get_stu_status(stu_id=stu_id)


class CDGK_MULTI(GDBaseModel):
    def build_cfg(self):
        self.n_user = self.datatpl_cfg['dt_info']['stu_count']
        self.n_item = self.datatpl_cfg['dt_info']['exer_count']
        self.n_cpt = self.datatpl_cfg['dt_info']['cpt_count']
        self.n_cpt_group = self.datatpl_cfg['dt_info']['n_cpt_group']


    def build_model(self):
        self.cdgk_list = nn.ModuleList(
            [
                CDGK_META(
                    stu_count=self.n_user, exer_count=self.n_item, cpt_count=self.n_cpt
                ) for _ in range(self.n_cpt_group)
            ]
        )
        for cdgk_meta in self.cdgk_list:
            cdgk_meta.Q_mat = self.Q_mat

    def add_extra_data(self, **kwargs):
        super().add_extra_data(**kwargs)
        self.Q_mat = kwargs['Q_mat'].to(self.device)

        self.gid2exers: Dict[int, torch.LongTensor] = {
            k: torch.from_numpy(v).long().to(self.device) 
            for k,v in kwargs['gid2exers'].items()
        }

        self.n_group_of_cpt: torch.LongTensor = torch.from_numpy(
            kwargs['n_group_of_cpt']
        ).long().to(self.device) 
        
    def forward(self, stu_id, exer_id, **kwargs):
        g_pd_list, g_order_list = [], []
        for gid, sub_cdgk in enumerate(self.cdgk_list):
            g_flag = torch.isin(exer_id, self.gid2exers[gid])
            if len(g_flag) == 0: continue

            g_stu_id = stu_id[g_flag]
            g_exer_id = exer_id[g_flag]
            g_pd = sub_cdgk(g_stu_id, g_exer_id)
            
            g_pd_list.append(g_pd)
            g_order_list.append(torch.argwhere(g_flag).flatten())

        order = torch.concat(g_order_list)
        pd = torch.concat(g_pd_list)

        pd, _ = self.groupby_mean(pd, labels=order, device=self.device)
        return pd.flatten()
    
    @torch.no_grad()
    def predict(self, **kwargs):
        return {"y_pd": self(**kwargs)}

    def get_main_loss(self, **kwargs):
        stu_id = kwargs['stu_id']
        exer_id = kwargs['exer_id']
        label = kwargs['label']
        pd = self(stu_id, exer_id).flatten()
        loss = F.binary_cross_entropy(input=pd, target=label)
        return {
            'loss_main': loss
        }

    def get_loss_dict(self, **kwargs):
        return self.get_main_loss(**kwargs)

    def get_stu_status(self, stu_id=None):
        if stu_id is not None:
            stu_emb_mat = torch.stack([sub_cdgk.stu_emb(stu_id) for sub_cdgk in self.cdgk_list], dim=2)
        else:
            stu_emb_mat = torch.stack([sub_cdgk.stu_emb.weight for sub_cdgk in self.cdgk_list], dim=2)
        
        return stu_emb_mat.sum(dim=-1) / self.n_group_of_cpt

    # reference: https://discuss.pytorch.org/t/groupby-aggregate-mean-in-pytorch/45335/9
    @staticmethod
    def groupby_mean(value:torch.Tensor, labels:torch.LongTensor, device="cuda:0"):
        """Group-wise average for (sparse) grouped tensors
        
        Args:
            value (torch.Tensor): values to average (# samples, latent dimension)
            labels (torch.LongTensor): labels for embedding parameters (# samples,)
        
        Returns: 
            result (torch.Tensor): (# unique labels, latent dimension)
            new_labels (torch.LongTensor): (# unique labels,)
            
        Examples:
            >>> samples = torch.Tensor([
                                [0.15, 0.15, 0.15],    #-> group / class 1
                                [0.2, 0.2, 0.2],    #-> group / class 3
                                [0.4, 0.4, 0.4],    #-> group / class 3
                                [0.0, 0.0, 0.0]     #-> group / class 0
                        ])
            >>> labels = torch.LongTensor([1, 5, 5, 0])
            >>> result, new_labels = groupby_mean(samples, labels)
            
            >>> result
            tensor([[0.0000, 0.0000, 0.0000],
                [0.1500, 0.1500, 0.1500],
                [0.3000, 0.3000, 0.3000]])
                
            >>> new_labels
            tensor([0, 1, 5])
        """
        uniques = labels.unique().tolist()
        labels = labels.tolist()

        key_val = {key: val for key, val in zip(uniques, range(len(uniques)))}
        val_key = {val: key for key, val in zip(uniques, range(len(uniques)))}
        
        labels = torch.LongTensor(list(map(key_val.get, labels)))
        
        labels = labels.view(labels.size(0), 1).expand(-1, value.size(1)).to(device)
        
        unique_labels, labels_count = labels.unique(dim=0, return_counts=True)
        result = torch.zeros_like(unique_labels, dtype=torch.float, device=device).scatter_add_(0, labels, value)
        result = result / labels_count.float().unsqueeze(1)
        new_labels = torch.LongTensor(list(map(val_key.get, unique_labels[:, 0].tolist())))
        return result, new_labels



import pandas as pd
from .raw2mid import BaseRaw2Mid

r"""
R2M_ASSIST_1516
########################
"""


class R2M_ASSIST_1516(BaseRaw2Mid):
    """R2M_ASSIST_1516 is a class used to handle the ASSISTment 2015-2016 dataset."""

    def process(self):
        super().process()
        pd.set_option("mode.chained_assignment", None)  # ignore warning
        # 读取原始数据，查看其属性
        raw_data = pd.read_csv(f"{self.rawpath}/2015_100_skill_builders_main_problems.csv", encoding='utf-8')

        # 获取交互信息
        # 对log_id进行排序，确定order序列
        inter = pd.DataFrame.copy(raw_data).sort_values(by='log_id', ascending=True)
        inter['order'] = range(len(inter))
        inter['label'] = inter['correct']
        df_inter = inter.rename(columns={'sequence_id': 'exer_id', 'user_id': 'stu_id'}).reindex(
            columns=['stu_id', 'exer_id', 'label', 'order', ]).rename(
            columns={'stu_id': 'stu_id:token', 'exer_id': 'exer_id:token', 'label': 'label:float',
                     'order': 'order_id:float'})

        # 获取学生信息
        stu = pd.DataFrame(set(raw_data['user_id']), columns=['stu_id', ])
        # stu['classId'] = None
        # stu['gender'] = None
        df_stu = stu.sort_values(by='stu_id', ascending=True)

        # 获取习题信息
        exer = pd.DataFrame(set(raw_data['sequence_id']), columns=['exer_id'])
        # exer['cpt_seq'] = None
        # exer['assignment_id'] = None
        df_exer = exer.sort_values(by='exer_id', ascending=True)

        # 此处将数据保存到`self.midpath`中

        df_inter.to_csv(f"{self.midpath}/{self.dt}.inter.csv", index=False, encoding='utf-8')
        # df_stu.to_csv(f"{self.midpath}/{self.dt}.stu.csv", index=False, encoding='utf-8')
        # df_exer.to_csv(f"{self.midpath}/{self.dt}.exer.csv", index=False, encoding='utf-8')
        pd.set_option("mode.chained_assignment", "warn")  # ignore warning
        return
